//////////////////////////   chess.h  //////////////////////////////////

#include <iostream>
//#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <fstream>

//#define NDEBUG
#include <assert.h>
//#define debug
using namespace std;

#ifndef TYPES
#define TYPES

static bool XBoard = false;

typedef unsigned long long i64;

enum piece_type {KING, QUEEN, ROOK, BISHOP, KNIGHT, PAWN, EMPTY};
enum color {WHITE, BLACK, NONE};
enum hash_entry_code {FAIL_LOW_UPPER, FAIL_HIGH_LOWER, EXACT_PV};
enum move_type{CAPTURE, ENPASSANT, PROMOXQ, PROMO2Q, CASTLEK, CASTLEQ, 
	NORMAL, SINGLE, PPUSH2, PROMOXR, PROMOXB, PROMOXN, PROMO2R, PROMO2B, PROMO2N, 
      NOMOVE}; 
//enum {HASHMOVE, GENWIN, WINNERS, KILLER1, KILLER2, GENHIST, HISTORIC};
//enum {MATERIAL, PIECES, BISHOP_PAIR, BISHOP_VS_KNIGHT, PAWN_SHIELD, PRESSURE, 
//      DEVELOPMENT, OUTPOSTS, PASSED, CONNECTED, CONTROL, PCSQ, NUMFEATURES};

enum piece_value {PAWN_VAL=100,KNIGHT_VAL=325,BISHOP_VAL=325,ROOK_VAL=500,QUEEN_VAL=975};

#define MAXPLY 64
#define MAXHIST 640
#define MAXMOVELIST 3072
#define MAINHASHSIZE 1048575   //entries --> must be (2^k)-1 for some integer k
//entries 500k = 524287  4194303 = 4m  1048575 = 1m[20bits]  2097151 = 2m
#define PAWNHASHSIZE 65535 //entries --> must be (2^k)-1 for some integer k
#define MATE 10000 // at root, ply = 1 so mate == MATE+(ply-1)
#define NFINITY 10000 // 9999 //10001 // 
#define OFFBOARD 0x88
#define CAPTURED 16

#define ASPIRWIN 40
// 975 + 1000 + 1300 = 3275   ( + 325 = 3600 )
#define MAXMAT 64 //3475 //  3275 //3600 // with queen extra 325 
#define MAXFORCE 34
#define MIN(x,y) (((x)<(y))?(x):(y))
#define MAX(x,y) (((x)>(y))?(x):(y))
#define ABS(x) abs(x)
//#define ABS(x) (((x)>0)?(x):-(x))

#define P2BIT(p,c) ((0x00010001 & (0x00008001 << (c))) << (p))
//#define P2BIT(p,c) ((0x00000001 << (c << 4)) << p)
//#define P2BIT(p,c) (0x00000001 << ((c) * 16 + (p)))

//#define HASHMATE(x) (x-2)

//#define MATED(x) (MATE+(x-1))
//#define WILLMATE(x) (
#define ISMATE(x) (x < -MATE+300 || x > MATE-300)

//#define TWO(c)     (0x1u << (c))
//#define MASK(c)    (((unsigned int)(-1)) / (TWO(TWO(c)) + 1u))
//#define COUNT(x,c) ((x) & MASK(c)) + (((x) >> (TWO(c))) & MASK(c))

/***************************** Structures **************************/

//struct sSTATUS {
//    int phase;
//    int mcount;
//};

//struct sFEATURE {
//    int weight;
//    int value;
//};

/************************* Class Declarations *************************/

class cMOVE {
public:    
    unsigned char from;
    unsigned char to;
    unsigned char mtype;
    int score;


    cMOVE() : from(OFFBOARD), to(OFFBOARD), mtype(NOMOVE), score(0)
    {}
    void set(int start, int dest, int typ, int sco){
        from = start;
        to = dest;
        mtype = typ;
        score = sco;
    }
    bool is_equal(cMOVE * m){
        if(to != m->to) return false;
        if(from != m->from) return false;
        if(mtype != m->mtype) return false;
        return true;
    }
    int is_pseudo_evasion();
	int is_volitile();
    bool is_pseudo();
    void to_string(char *);
	int to_san(char * san);
    void print();
	
};

class cSQUARE {
public:
    int color;
    int ptype;
    int pindex;
    
    cSQUARE() : color(NONE), ptype(EMPTY), pindex(CAPTURED)//, affected(0)
    { }
    void put(int col, int typ, int index){
        color = col;
        ptype = typ;
        pindex = index;
    }
    void empty(){
        color = NONE;
        ptype = EMPTY;
        pindex = CAPTURE;
    }
    int iswhitepawn(){
        return (ptype == PAWN && color == WHITE);
    }
    int isblackpawn(){
        return (ptype == PAWN && color == BLACK);
    }
	int ispawnofcolor(int c){
		return (ptype == PAWN && color == c);
	}
	int isempty(){
		return (ptype == EMPTY);
	}
	int isoccupied(){
		return (ptype != EMPTY);
	}
	int iscolor(int c){
		return (color == c);
	}
};

class cHISTENTRY {
    cMOVE move;
    cSQUARE victim;
	int fifty;
	i64 key;
    unsigned int affected;
public:
	int castle;
    int enpassant;
    int matscore;
    friend class cGAMEHIST;
};

class cGAMEHIST {
    
public:
    cHISTENTRY stack[MAXHIST];
	int hply;
    cGAMEHIST(): hply(1)
    {}
    void push(cMOVE *, cSQUARE, unsigned int);
    void pop(cMOVE &, cSQUARE &, unsigned int &);
    void pushnull();
    void popnull();
    bool was_recap();
	bool was_p27();
	cMOVE * top();
    int last_victim();
    bool rep();
	bool rep2();
	void print();
	int progress(int);
};

class cBOARD {
public:    
    cGAMEHIST gamestack;
    cSQUARE square[128];
    unsigned int control[128];
    int pieces[2][16];
    
    int material[2]; 
    int advantage[2];   
    int liability[2];
    
    int piece_count[2][6];

	int majors[2];
	int minors[2];

    int least_adv_pawn[2][10];
    int most_adv_pawn[2][10];
	unsigned char wpasser, bpasser;
    //int lat_onfile[2][10];
    
	int hits[2];
	//int esquares[2];
    int proximity[2];
	int defenders[2];
    int undeveloped[2];
    //int couldcastle;
    
    int smallest[2];
    int largest[2];
    
    unsigned int attack_mask[2][6];
    unsigned int minor_attack[2];
    unsigned int major_attack[2];
    unsigned int has_pieces[2];
    
	bool incheck();

	int null_ok();

	//int eval(int, int);
    void pre_eval();
    int evaluate();
    void eval_pieces();
    int find_pawn();
    int eval_pawn_race(int);
    void eval_king_safety();
    int seval();
    
    int count_ctrl();
    void setup(char *);
	void flip();
    void buildepd(char *);
    int domove(cMOVE *);
    void unmove();
    void display();
    void setaffected(int, int, int, unsigned int);
    void clraffected(int, int, int, unsigned int);
    void clear_all(unsigned int);
    void set_all(unsigned int);
    i64 hash();
    int isok();
    void init_affected();
    
};

class cMOVELIST {
public:
    cMOVE hashmove[MAXPLY];
    cMOVE killer1[MAXPLY];
    cMOVE killer2[MAXPLY];
    cMOVE moves[MAXMOVELIST];
    
    cMOVE * nextmove[MAXPLY];
    cMOVELIST(){
        nextmove[0] = moves;
    }
    
    int mode[MAXPLY];
    cMOVE * next();
    
    void gen();
    void gencap();
	//void genquiet();
	//void gencap2();
	int filterlegals();
    //void bsortall();
    void isortall();
    void isortone(cMOVE *);
    void list();
    /*void genwk(int);
    void genbk(int);
    void genwq(int);
    void genbq(int);
    void genwr(int);
    void genbr(int);
    void genwb(int);
    void genbb(int);
    void genwn(int);
    void genbn(int);
    void genwp(int);
    void genbp(int);*/
    void addmove(int, int, int, int);
    /*void gencapwk(int);
    void gencapbk(int);
    void gencapwq(int);
    void gencapbq(int);
    void gencapwr(int);
    void gencapbr(int);
    void gencapwb(int);
    void gencapbb(int);
    void gencapwn(int);
    void gencapbn(int);
    void gencapwp(int);
    void gencapbp(int);*/
};     

struct sMOVE {
    unsigned char from;
    unsigned char to;
    unsigned char type;
};
  
class cHASHENTRY {
public:    
    i64 lock; //  64  [minimum 
    short int value;  // 16   [minimum 15 bits  14 if use NFINITY = 8191
    char depth; // 8 bits [min 5 or 6
    char data; /* 76543210       8 bits  min 5
                  ......xx  type
                  ...xxx--  id  */
    sMOVE bm; // 24 bits  min 18
};

class cPHASHENTRY {
public:
    unsigned int lock;
    short int wliability, bliability;
	short int wpps, bpps;
	unsigned char wpassed; // bit indicates passed pawns on file
	unsigned char bpassed;
};

class cPHASHTABLE {
    cPHASHENTRY *table;
    
public:
	size_t pawn_hash_table_size;

	cPHASHTABLE(size_t size) {
		pawn_hash_table_size = PAWNHASHSIZE;
		size_t test = size;
		// We check to see that every bit starting from the lowest up to the largest bit set is on:
		while (test)
		{
			if ((test & 1) == 0) {
				puts("Error in hash table size (must be of the form (2^k)-1) for some k.");
				puts("Using default pawn hash size.");
				size = PAWNHASHSIZE;
				break;
			}
            test >>= 1;
		}
		
		table = new cPHASHENTRY[size+1];
		pawn_hash_table_size = size;
	}

	cPHASHTABLE() {
		pawn_hash_table_size = PAWNHASHSIZE;
		table = new cPHASHENTRY[pawn_hash_table_size+1];
	}
	~cPHASHTABLE() {
		delete [] table;
	}

	void resize(size_t size) {
		if (table) {
			delete [] table;
			table = NULL;
		}

		pawn_hash_table_size = PAWNHASHSIZE;
		size_t test = size;
		// We check to see that every bit starting from the lowest up to the largest bit set is on:
		while (test)
		{
			if ((test & 1) == 0) {
				puts("Error in hash table size (must be of the form (2^k)-1) for some k.");
				puts("Using default pawn hash size.");
				size = PAWNHASHSIZE;
				break;
			}
            test >>= 1;
		}
		//cout << "pawn table size = " << size << endl;
		table = new cPHASHENTRY[size+1];
		pawn_hash_table_size = size;
		clear();
	}

    void clear();
    //void store(int val);
    cPHASHENTRY * probe(unsigned int);
};

i64 rand64();

class cHASHTABLE {
    cHASHENTRY *table;

public:
    int age;
    int stored;
    int probed;
    int matched;
    int useful;
    int collisions;
	size_t main_hash_size;

	cHASHTABLE(size_t size) {
		main_hash_size = MAINHASHSIZE;
		size_t test = size;
		// We check to see that every bit starting from the lowest up to the largest bit set is on:
		while (test)
		{
			if ((test & 1) == 0) {
				puts("Error in hash table size (must be of the form (2^k)-1) for some k.");
				puts("Using default main hash size.");
				size = MAINHASHSIZE;
				break;
			}
            test >>= 1;
		}
		table = new cHASHENTRY[size+1];
		main_hash_size = size;
	}

    cHASHTABLE(){
		main_hash_size = MAINHASHSIZE; 
		table = new cHASHENTRY[main_hash_size+1];
        age = 4;
        stored = 0;
        collisions = 0;
        probed = 0;
        matched = 0;
        useful = 0;
    }
    ~cHASHTABLE(){
		delete [] table;
	}

	void resize(size_t size) {
		if (table)
		{
		    delete [] table;
			table = NULL;
		}

		main_hash_size = MAINHASHSIZE;
		size_t test = size;
		// We check to see that every bit starting from the lowest up to the largest bit set is on:
		while (test)
		{
			if ((test & 1) == 0) {
				puts("Error in hash table size (must be of the form (2^k)-1) for some k.");
				puts("Using default main hash size.");
				size = MAINHASHSIZE;
				break;
			}
            test >>= 1;
		}
		//cout << "main table size = " << size << endl;
		table = new cHASHENTRY[size+1];
		main_hash_size = size;
		clear();
	}


    void clear();
    int count();
    void inc_age();
    void store(int, int, int, cMOVE *);
    void stuffmove(cMOVE *);
    cHASHENTRY * probe();
};


/*****************************  Prototypes ******************************/
int perft(int);
void init_data();
int Bioskey();
void clear_historic();
void age_historic();
bool s2move(char* s, cMOVE &m);
//char * GetBookM();
cMOVE think(int, int, int, char*);
//void iterate(int, int);
void bump(cMOVE *, int);
void stuff(int, int);
void extract(cMOVE *);
int pvsearch(int, int, int, int, int);
int csearch(int, int);
int isearch(int, int);
int qsearch(int, int);
int run_suite(char *, int , int);
int eval_suite(char *);
void winboard();
unsigned int popcount16(unsigned int);
unsigned int BSR(unsigned int);
unsigned long genrand();
int onking(int);
int mate_search(int, int, int);
int read_ini_file(char *);

#endif     
