
#ifdef _LINUX
#include <stdlib.h>
#define ABS abs
#endif

#include "chess.h"
#include "data.h"
#include "weights.h"
#include "search.h"


/*
    capture search
	captures, promotions[Q ONLY] and Check evasions alone!
	will detect mate
	returns a minimum of stand-pat static eval unless mated
*/

int csearch(int alpha, int beta){
    
	//int temp = 0;
	//assert((temp = board.isok())==0);
	assert(board.isok()==0);
	
	//qnodes++;

#ifdef TRACEc
	logfile << key << "\t\tENTc\t" << alpha << "\t" << beta << "\t" << 0 << "\t" << ply << endl;
#endif

	if(ply > qdepth) qdepth = ply;
	if(ply >= MAXPLY-2) return beta;
    pvl[ply] = ply;
    
	if(fifty>3){ 
        if(board.gamestack.rep()) return 0;
		if(fifty >= move50) return 0;
        
	}

	// time check.  @100knps time check every 1/12 sec
    if((qnodes & 8191) == 8191)
        if(clock() >= st /* || Bioskey() */){
            stopsearch = 1;
            return 0;
        }

	cMOVE bm;
    int val;
    
	/*
	next ply killer is blank for 1st move, but >1 have a killer
	ie, we try first move and recurse into search but have no killer there
	but upon searching consecutive moves, hopefully we found killer
	*/

    tree.killer1[ply+1].to = 0x88; 
	tree.killer2[ply+1].to = 0x88; 
	tree.hashmove[ply].to = 0x88;

#ifdef HASHING	

	cHASHENTRY * he = hashtable.probe();
    if(he != NULL){
        tree.hashmove[ply].from = he->bm.from;
        tree.hashmove[ply].to = he->bm.to;
        tree.hashmove[ply].mtype = he->bm.type;
        val = he->value;
		
		if(ABS(val) > MATE-300){
            if(val > 0) val -= (ply-0);
            else val += (ply-0);
        }

        switch(he->data & 3){
            case FAIL_LOW_UPPER: // All searched. val has best score
				if(val <= alpha){
					hashtable.stored += ((he->data & 60) != hashtable.age);
					return val;
				}
                break; 
            case FAIL_HIGH_LOWER: // Few searched. val has best
                if(val >= beta){
					hashtable.stored += ((he->data & 60) != hashtable.age);
					return val;
				}
                break;
            case EXACT_PV:
                if(val > alpha && val < beta)
                    extract(&tree.hashmove[ply]);
                hashtable.stored += ((he->data & 60) != hashtable.age);
				return val;
                break;
        }
    }
	
#endif

	bm = tree.hashmove[ply];

	bool incheck = board.incheck();
	int nt = FAIL_LOW_UPPER;

	if(incheck) val = -MATE +(ply-1); // + 2;//-NFINITY;//
	else val = board.evaluate();

	if(val > alpha){
		if(val >= beta){
			hashtable.store(0, val, FAIL_HIGH_LOWER, &bm);

#ifdef TRACEc
			logfile << key << "\t\tcSTP\t" << alpha << "\t" << beta << "\t" << val << "\t" << 0 << "\t" << ply << endl;
#endif				
			return val;
		}
		alpha = val;
	}


	if(incheck) tree.gen();
	else tree.gencap();

	int legal = 0, stdpat = val, best = -NFINITY;;
	
	
	cMOVE * cm = tree.nextmove[ply-1];
    while(cm < tree.nextmove[ply]){
        tree.isortone(cm);

		if(!incheck // only caps were generated, so they're all 'good'
			|| cm->is_pseudo_evasion() // consider only moves which may evade
		){ 
			
#ifdef TRACEc
			i64 lock = key;
			cm->to_san(buffer);
#endif
			
			if(board.domove(cm)){
				legal++;

				/*
					if we're in check and searching for evasions
					or this capture gives check _AND_ could be mate, 
					or this is a non 'futile' capture then search it,
					otherwise - prune it
				*/
#ifdef FUTILITYc				
				if(incheck 
					|| board.incheck()

					|| stdpat + board.gamestack.last_victim() + max_pos_gain > alpha

					){

#ifdef TRACEc
	logfile << lock << "\t" << buffer << "\tCSE\t" << -alpha << "\t" << -beta << "\t" << 0 << "\t" << ply << endl;;
#endif
					qnodes++;
					val = -csearch(-beta, -alpha);
				}
#else
				qnodes++;
				val = -csearch(-beta, -alpha);
#endif				
				board.unmove();
				if(val > best){
					if(val > alpha){   
						if(val >= beta){  
							if(cm->score < 128) bump(cm, 1);
							fh1 += (legal == 1);
							fhn++;
							hashtable.store(0, val, FAIL_HIGH_LOWER, cm);
#ifdef TRACEc
							cm->to_san(buffer);
							logfile << key << "\t" << buffer << "\tcCUT\t" << alpha << "\t" << beta << "\t" << val << "\t" << 0 << "\t" << ply << endl;
#endif
							return val;
						}
						alpha = val;
						//bm = *cm;
						nt = EXACT_PV;
						/* update the PV */
						pv[ply][ply] = *cm;
						for (int j = ply + 1; j < pvl[ply + 1]; ++j)
							pv[ply][j] = pv[ply + 1][j];
						pvl[ply] = pvl[ply + 1];
					}
					bm = *cm;
					best = val;
					if(best > stdpat) stdpat = best;
				}
			}
		}
        cm++;
    }
	
	/*
		we are iether mated, stalemated, but since only caps are searched when
		NOT in check, we dont really know if we're stalemated
	*/

	if(legal==0){
		if(incheck){
			stdpat = -MATE+(ply-1);
			nt = EXACT_PV;
#ifdef TRACEc
			logfile << key << "\t*********** cMATE ***********\n";
#endif
			/*board.display();
			cout << "\ncsearch()\n";
			getchar();*/
		}//else if(stdpat <= -MATE+300) cout << "danger Will\n";
		//hashtable.store(0, stdpat, nt, &no_move);
		//return stdpat;
	}
    
	// non-cap / losing-cap pv only
	//if(nt == EXACT_PV && bm.score < 128) bump(&bm, 1);

    hashtable.store(0, stdpat, nt, &bm);

#ifdef TRACEc
	bm.to_san(buffer);
	if(nt == EXACT_PV)
		logfile << key << "\t" << buffer << "\tcPVN\t" << alpha << "\t" << beta << "\t" << stdpat << "\t" << 0 << "\t" << ply << endl;
	else
		logfile << key << "\t" << buffer << "\tcALL\t" << alpha << "\t" << beta << "\t" << stdpat << "\t" << 0 << "\t" << ply << endl;
#endif	
	
	//assert(stdpat > -MATE && stdpat < MATE);

	return stdpat;
}

