#include "chess.h"
#include "data.h"

#ifdef _LINUX
#include <stdlib.h>
#endif


void GetBookM(char *);
void playself(int games){
	cMOVE m;
	char line[8];
	ofstream pgn;
	pgn.open("games.pgn");
	int t, d;
	do{
		board.setup("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq");
		pgn << "[Event \"Self Play Test\"]" << endl
		<< "[Site \"AMDATHLON\"]" << endl
		<< "[Date \"????.??.??\"]" << endl
		<< "[Round \"?\"]" << endl
		<< "[White \"GES\"]" << endl
		<< "[Black \"GES\"]" << endl
		<< "[Result \"\"]" << endl
		<< "[ECO \"\"]" << endl 
		<< endl;

		t = 3;
		d = 0;
		do{
			system("cls");
			board.display();
			
			if(board.gamestack.hply < 40){
				GetBookM(line);
				if(line == NULL || line[0] == 0) m = think(6000,0,1,NULL); // sometimes getmove chooses not to
				else if(s2move(line, m));
				else{
					cout << "\aBad Book Move! " << line << endl;
					return;
				}
			}
			else m = think(6000,0,1,NULL);
			
			if(m.from == 0x88 || m.to == 0x88){
				cout << "\aBad move from think ";
				m.print();
				return;
			}
			
			if(m.score < -600 || m.score > 600) break;
			
			if(board.gamestack.hply > 39){
				if(m.score == 0){
					d++;
					if(d > 10) break;
				}else d = 0;
				// this may make the above redundant
				if(board.gamestack.progress(m.score) < 11){
					m.score = 0;
					break;
				}
			}

			//game not over and move appears to be ok
			if(t&1) pgn << (t >> 1) << ". ";
			m.to_san(line);
			pgn << line << " ";
			if(((t-3)&7) == 7) pgn << endl;
			
			if(board.domove(&m)){
				tree.gen();
				t++;
			}
			else break;
			
		}while(1);

		if(m.to & 0x88)
			pgn << "*";
		else{
			if(side == WHITE){
				if(m.score < 0) pgn << "0-1";
				else if(m.score > 0) pgn << "1-0";
				else pgn << "=-=";
			}
			else{
				if(m.score < 0) pgn << "1-0";
				else if(m.score > 0) pgn << "0-1";
				else pgn << "=-=";
			}
		}
		
		pgn << endl << endl << endl;

	}while(--games > 0);
	pgn.close();
}


void cGAMEHIST::print(){
	for(int i = 1; i < hply; i++){
		cout << i
			<< "\t" << stack[i].key
			<< "\t" << stack[i].fifty
			<< "\t" << stack[i].castle
			<< "\t" << stack[i].enpassant << endl;
	}
}

i64 cBOARD::hash(){
    i64 key = 0;
    
	int i = 0;
    do{
        if(square[i].ptype != EMPTY)
            key ^= zobrist[square[i].color][square[i].ptype][i];
        if(i == 119) break;
        if((i & 7)==7) i += 9;
        else i++;
    }while(1);
	
	key ^= zcastle[castle & 15];
    key ^= zenpassant[enpassant];
    
    if(side == BLACK) key = ~key;
    return key;
}

int cBOARD::isok(){
    int p, c, f;
    int pc[2] = {0, 0};
    int kn[2] = {0, 0};
	int ma[2] = {0, 0};
	int no[2] = {0, 0};
    int sq;
	
	if(key != hash()) return 11;

	for(c = 0; c < 2; c++)
		for(p = 0; p < 16; p++){
			if((sq=pieces[c][p])&0x88){ // piece is off board and not flagged 'CAPTURED'
				if(sq != 0x88) return 1;
			}
			else{
				if(square[sq].color != c) return 2;
				if(square[sq].pindex != p) return 3;
				if(square[sq].ptype == EMPTY) return 4;
			}
		}

    sq = 0;
	do{
        if(sq&0x88);
        else{
            p = square[sq].ptype;
            c = square[sq].color;
            if(p == KING){
                kn[c]++;
                if(kn[c] > 1) return 5;
            }
            else if(p != EMPTY){
                pc[c]++;
                if(pc[c] > 15) return 6;
				ma[c] += matval[p];
				no[c] += forceval[p];
				// search c piece list for p at sq
				f = 0;
				for(int i = 1; i < 16; i++)
					if(pieces[c][i] == sq){f = 1; break;}
				if(!f) return 7;
            }
        }
    }while(++sq < 120);
	if(ma[0] != material[0] || ma[1] != material[1]) return 8;
	if(no[0] != force[0] || no[1] != force[1]) return 9;
    if(kn[0] & kn[1]) return 0;
    return 10;
}

void cBOARD::flip(){
	int t1,t2;
	cSQUARE t;
	//puts("flipping board ...\n");
	for(int j = 0; j < 4; j++){
		for(int i = 0; i < 8; i++){
			t = square[j*16 + i];
			square[16 * j + i] = square[16 * (7-j) + i];
			square[16 * (7-j) + i] = t;
			if(square[16 * j + i].ptype != EMPTY) square[16 * j + i].color ^= 1;
			if(square[16 * (7-j) + i].ptype != EMPTY) square[16 * (7-j) + i].color ^= 1;
		}
	}
	//puts("adjusting castle rights and enpassant sq ...\n");
	t1 = castle & 15;
	t2 = t1 >> 2;
	t1 <<= 2;
	t1 &= 15;
	t1 |= t2; // has castle rights

	castle >>= 4;
	t2 = castle & 3;
	t2 <<= 2;
	castle >>= 2;
	castle |= t2;
	castle <<= 4; // put nible back on left
	castle |= t1;

	enpassant = (enpassant & 7) + ((7 - (enpassant >> 4)) << 4);

	side ^= 1;
	xside ^= 1;
	
	buildepd(buffer);
	setup(buffer);
}

