#include "chess.h"
#include "data.h"

int cBOARD::domove(cMOVE * move){
    int from = move->from;
    int dest = move->to;
    int color = square[from].color;
    int other = color ^ 1;
    int pindex = square[from].pindex;
    int ptype = square[from].ptype;
    
	int castleo = castle;
	int enpassanto = enpassant;

#ifdef debug
    if((from & 0x88) || (dest & 0x88)) cout << "movement off board  ";
    if(color > 1) cout << "moving colorless piece  ";
    if(pindex > 15) cout << "moving captured piece  ";
    if(ptype > 5) cout << "moving unknown piece  ";
    if(ply >= MAXPLY-1) cout << "Too deep  ";
#endif

// CAPTURE, ENPASSANT, PROMOXQ,  
// PROMO2Q,  
// CASTLEK, CASTLEQ,
// NORMAL, SINGLE, PPUSH2,     
// PROMOXR, PROMOXB, PROMOXN,
// PROMO2R, PROMO2B, PROMO2N,
// NOMOVE

/*
	Too often, a bit is set in affected using P2BIT only to be extracted over in control
	when it would have been simpler to send pindex
	We still may need P2BIT however that it may be excluded from affected
*/

    cSQUARE victim;
    int vsq; // victim square
    unsigned int vbit; // victim bit -for affected
    unsigned int affected = control[from];// | control[dest]; // only for xray
    switch(move->mtype){
        case CAPTURE:
            victim = square[dest];
            vbit = P2BIT(victim.pindex, other);
            affected |= P2BIT(pindex, color);
            //clraffected(from, color, ptype, P2BIT(pindex, color));
            clear_all(affected | vbit);
            square[dest] = square[from];
            square[from].empty(); // empty
            pieces[other][victim.pindex] = 0x88;
            pieces[color][pindex] = dest;  // change address
            //setaffected(dest, color, ptype, P2BIT(pindex, color));
            set_all(affected & ~vbit & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                //clraffected(dest, color, ptype, P2BIT(pindex, color));
                clear_all(affected & ~vbit & mask[color]);
                square[from] = square[dest];
                square[dest] = victim;
                pieces[color][pindex] = from; // return to original address
                pieces[other][victim.pindex] = dest;
                //setaffected(from, color, ptype, P2BIT(pindex, color));
                set_all(affected | vbit);
                return 0;
            }
            // but own piece could be affected by own move and would not
            // clear all affected
            //clraffected(from, color, ptype, P2BIT(pindex, color));
            //setaffected(dest, color, ptype, P2BIT(pindex, color));
            set_all(affected /*& ~vbit*/ & mask[other]);//vbit excluded already by mask
            gamestack.push(move, victim, affected | vbit);
            fifty = 0;
            enpassant = 0;
            castle &= (castlemask[from] & castlemask[dest]);
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[color][ptype][dest];
            key ^= zobrist[other][victim.ptype][dest];
            material[other] -= matval[victim.ptype];
			force[other] -= forceval[victim.ptype];
            break;
//---------------------------------------------------- 
        case ENPASSANT:
            /* figure out what is affected */
            vsq = (color == WHITE)? dest+16 : dest-16;
			//vsq = dest + 16 - (color << 5);
            victim = square[vsq];

            affected |= control[vsq] | control[dest];
            vbit = P2BIT(victim.pindex, other);
            affected |= P2BIT(pindex, color);
            clear_all(affected | vbit);
            /*  move the piece on board and change address */
            square[dest] = square[from];
            square[from].empty();
            pieces[color][pindex] = dest;
            square[vsq].empty();
            pieces[other][victim.pindex] = 0x88;
            set_all(affected & ~vbit & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                clear_all(affected & ~vbit & mask[color]);
                square[from] = square[dest];
                square[vsq] = victim;
                square[dest].empty();
                pieces[color][pindex] = from; // return to original address
                pieces[other][victim.pindex] = vsq;
                set_all(affected | vbit);
                return 0;
            }
            set_all(affected /*& ~vbit*/ & mask[other]);
            gamestack.push(move, victim, affected | vbit);
            fifty = 0;
            enpassant = 0;
            // no way this affects castle
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[color][ptype][dest];
            key ^= zobrist[other][PAWN][vsq];
            material[other] -= matval[victim.ptype];
            break;
//----------------------------------------------------
        case PROMOXQ:
            victim =square[dest];
            vbit = P2BIT(victim.pindex, other);
            affected |= P2BIT(pindex, color);
            clear_all(affected | vbit);
            square[dest].put(color, QUEEN, pindex);
            square[from].empty(); // empty
            pieces[other][victim.pindex] = 0x88;
            pieces[color][pindex] = dest;  // change address
            set_all(affected & ~vbit & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                clear_all(affected & ~vbit & mask[color]);
                square[from].put(color, PAWN, pindex);
                square[dest] = victim;
                pieces[color][pindex] = from; // return to original address
                pieces[other][victim.pindex] = dest;
                set_all(affected | vbit);
                return 0;
            }
            set_all(affected /*& ~vbit*/ & mask[other]);
            gamestack.push(move, victim, affected | vbit);
            fifty = 0;
            enpassant = 0;
            castle &= (castlemask[from] & castlemask[dest]);
            key ^= zobrist[color][QUEEN][dest];
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[other][victim.ptype][dest];
            material[color] += QUEEN_VAL-PAWN_VAL;
            material[other] -= matval[victim.ptype];
            force[color] += forceval[QUEEN];
            force[other] -= forceval[victim.ptype];
            break;
//----------------------------------------------------
        case PROMO2Q:
            affected |= P2BIT(pindex, color)  | control[dest];
            clear_all(affected);
            square[dest].put(color, QUEEN, pindex);
            square[from] = victim; // empty
            pieces[color][pindex] = dest;  // change address
            set_all(affected & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                clear_all(affected & mask[color]);
                square[from].put(color, PAWN, pindex);
                square[dest] = victim; // empty
                pieces[color][pindex] = from; // return to original address
                set_all(affected);
                return 0;
            }
            set_all(affected & mask[other]);
            gamestack.push(move, victim, affected);
            fifty = 0;
            enpassant = 0;
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[color][QUEEN][dest];
            material[color] += QUEEN_VAL-PAWN_VAL;
            force[color] += forceval[QUEEN];
            break;
//----------------------------------------------------
        case CASTLEK: // if possible/legal just do it and exit
            affected |= P2BIT(pindex, color) | control[dest];
            if(side == WHITE){
               affected |= control[119] | control[117];
               //affected |= (1 << ( + ( << 4)));
               affected |= P2BIT(square[119].pindex, color);
               gamestack.push(move, victim, affected);
               clear_all(affected);
               //moveto(from, dest); // move king
               square[dest] = square[from];
               square[from] = victim; // empty
               pieces[color][pindex] = dest;  // change address
               key ^= zobrist[color][ptype][from];
               key ^= zobrist[color][ptype][dest];
               //moveto(119, 117);  // move rook
               square[117] = square[119];
               square[119] = victim; // empty
               pieces[color][square[117].pindex] = 117;  // change address
               key ^= zobrist[color][ROOK][119];
               key ^= zobrist[color][ROOK][117];
               castle &= 252;//12;
               castle |= 16;
            }
            else{
               affected |= control[7] | control[5];
               affected |= P2BIT(square[7].pindex, color);
               gamestack.push(move, victim, affected);
               clear_all(affected);
               //moveto(from, dest); // move king
               square[dest] = square[from];
               square[from] = victim; // empty
               pieces[color][pindex] = dest;  // change address
               key ^= zobrist[color][ptype][from];
               key ^= zobrist[color][ptype][dest];
               //moveto(7, 5); // move rook
               square[5] = square[7];
               square[7] = victim; // empty
               pieces[color][square[5].pindex] = 5;  // change address
               key ^= zobrist[color][ROOK][7];
               key ^= zobrist[color][ROOK][5];
               castle &= 243;//3;
               castle |= 64;
            }
            fifty = 0;
            enpassant = 0;
            set_all(affected);
            break;
//----------------------------------------------------
        case CASTLEQ: // if possible/legal just do it and exit
            affected |= P2BIT(pindex, color) | control[dest];
            if(side == WHITE){
               affected |= control[112] | control[115];
               affected |= P2BIT(square[112].pindex, color);
               gamestack.push(move, victim, affected);
               clear_all(affected);
               //moveto(from, dest); // move king
               square[dest] = square[from];
               square[from] = victim; // empty
               pieces[color][pindex] = dest;  // change address
               key ^= zobrist[color][ptype][from];
               key ^= zobrist[color][ptype][dest];
               //moveto(112, 115);
               square[115] = square[112];
               square[112] = victim; // empty
               pieces[color][square[115].pindex] = 115;  // change address
               key ^= zobrist[color][ROOK][112];
               key ^= zobrist[color][ROOK][115];
               castle &= 252;//12;
               castle |= 32;
            }
            else{
               affected |= control[0] | control[3];
               affected |= P2BIT(square[0].pindex, color);
               gamestack.push(move, victim, affected);
               clear_all(affected);
               //moveto(from, dest); // move king
               square[dest] = square[from];
               square[from] = victim; // empty
               pieces[color][pindex] = dest;  // change address
               key ^= zobrist[color][ptype][from];
               key ^= zobrist[color][ptype][dest];
               //moveto(0, 3);
               square[3] = square[0];
               square[0] = victim; // empty
               pieces[color][square[3].pindex] = 3;  // change address
               key ^= zobrist[color][ROOK][0];
               key ^= zobrist[color][ROOK][3];
               castle &= 243;//3;
               castle |= 128;
            }
            fifty = 0;
            enpassant = 0;
            set_all(affected);
            break;
//----------------------------------------------------
        case NORMAL:
            affected |= P2BIT(pindex, color) | control[dest];
            clear_all(affected);
            //moveto(from, dest);
            square[dest] = square[from];
            square[from] = victim; // empty
            pieces[color][pindex] = dest;  // change address
            set_all(affected & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                clear_all(affected & mask[color]);
                square[from] = square[dest];
                square[dest] = victim; // empty
                pieces[color][pindex] = from; // return to original address
                set_all(affected);
                return 0;
            }
            set_all(affected & mask[other]);
            gamestack.push(move, victim, affected);
            fifty++;
            enpassant = 0;
            castle &= (castlemask[from] & castlemask[dest]);
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[color][ptype][dest];
            break;
//----------------------------------------------------
        case SINGLE:   
            affected |= P2BIT(pindex, color) | control[dest];
            clear_all(affected);
            square[dest] = square[from];
            square[from] = victim; // empty
            pieces[color][pindex] = dest;  // change address
            set_all(affected & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                clear_all(affected & mask[color]);
                square[from] = square[dest];
                square[dest] = victim; // empty
                pieces[color][pindex] = from; // return to original address
                set_all(affected);
                return 0;
            }
            set_all(affected & mask[other]);
            gamestack.push(move, victim, affected);
            fifty = 0;
            enpassant = 0;
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[color][ptype][dest];
            break;
//---------------------------------------------------- 
        case PPUSH2:
            affected |= P2BIT(pindex, color) | control[dest];
            clear_all(affected);
            square[dest] = square[from];
            square[from] = victim; // empty
            pieces[color][pindex] = dest;  // change address
            set_all(affected & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                clear_all(affected & mask[color]);
                square[from] = square[dest];
                square[dest] = victim; // empty
                pieces[color][pindex] = from; // return to original address
                set_all(affected);
                return 0;
            }
            set_all(affected & mask[other]);
            gamestack.push(move, victim, affected);
            fifty = 0;
            if(color == WHITE){
                if(square[dest+1].isblackpawn())
                     enpassant = dest+16;
                else if(square[dest-1].isblackpawn())
                     enpassant = dest+16;
                else enpassant = 0;
            }
            else{
                if(square[dest+1].iswhitepawn())
                     enpassant = dest-16;
                else if(square[dest-1].iswhitepawn())
                     enpassant = dest-16;
                else enpassant = 0;
            }    
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[color][ptype][dest];
            break;
//----------------------------------------------------
        case PROMOXR:
            victim =square[dest];
            vbit = P2BIT(victim.pindex, other);
            affected |= P2BIT(pindex, color);
            clear_all(affected | vbit);
            square[dest].put(color, ROOK, pindex);
            square[from].empty(); // empty
            pieces[other][victim.pindex] = 0x88;
            pieces[color][pindex] = dest;  // change address
            set_all(affected & ~vbit & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                clear_all(affected & ~vbit & mask[color]);
                square[from].put(color, PAWN, pindex);
                square[dest] = victim;
                pieces[color][pindex] = from; // return to original address
                pieces[other][victim.pindex] = dest;
                set_all(affected | vbit);
                return 0;
            }
            set_all(affected /*& ~vbit*/ & mask[other]);
            gamestack.push(move, victim, affected | vbit);
            fifty = 0;
            enpassant = 0;
            castle &= (castlemask[from] & castlemask[dest]);
            key ^= zobrist[color][ROOK][dest];
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[other][victim.ptype][dest];
            material[color] += ROOK_VAL-PAWN_VAL;
            material[other] -= matval[victim.ptype];
            force[color] += 5;
            force[other] -= forceval[victim.ptype];
            break;
//----------------------------------------------------
        case PROMOXB:
            victim =square[dest];
            vbit = P2BIT(victim.pindex, other);
            affected |= P2BIT(pindex, color);
            clear_all(affected | vbit);
            square[dest].put(color, BISHOP, pindex);
            square[from].empty(); // empty
            pieces[other][victim.pindex] = 0x88;
            pieces[color][pindex] = dest;  // change address
            set_all(affected & ~vbit & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                clear_all(affected & ~vbit & mask[color]);
                square[from].put(color, PAWN, pindex);
                square[dest] = victim; // empty
                pieces[color][pindex] = from; // return to original address
                pieces[other][victim.pindex] = dest;
                set_all(affected | vbit);
                return 0;
            }
            set_all(affected /*& ~vbit*/ & mask[other]);
            gamestack.push(move, victim, affected | vbit);
            fifty = 0;
            enpassant = 0;
            castle &= (castlemask[from] & castlemask[dest]);
            key ^= zobrist[color][BISHOP][dest];
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[other][victim.ptype][dest];
            material[color] += BISHOP_VAL-PAWN_VAL;
            material[other] -= matval[victim.ptype];
            force[color] += 3;
            force[other] -= forceval[victim.ptype];
            break;
//----------------------------------------------------
        case PROMOXN:
            victim =square[dest];
            vbit = P2BIT(victim.pindex, other);
            affected |= P2BIT(pindex, color);
            clear_all(affected | vbit);
            square[dest].put(color, KNIGHT, pindex);
            square[from].empty(); // empty
            pieces[other][victim.pindex] = 0x88;
            pieces[color][pindex] = dest;  // change address
            set_all(affected & ~vbit & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                clear_all(affected & ~vbit & mask[color]);
                square[from].put(color, PAWN, pindex);
                square[dest] = victim; // empty
                pieces[color][pindex] = from; // return to original address
                pieces[other][victim.pindex] = dest;
                set_all(affected | vbit);
                return 0;
            }
            set_all(affected /*& ~vbit*/ & mask[other]);
            gamestack.push(move, victim, affected | vbit);
            fifty = 0;
            enpassant = 0;
            castle &= (castlemask[from] & castlemask[dest]);
            key ^= zobrist[color][KNIGHT][dest];
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[other][victim.ptype][dest];
            material[color] += KNIGHT_VAL-PAWN_VAL;
            material[other] -= matval[victim.ptype];
            force[color] += 3;
            force[other] -= forceval[victim.ptype];
            break;
//----------------------------------------------------
        case PROMO2R:
            affected |= P2BIT(pindex, color) | control[dest];
            clear_all(affected);
            square[dest].put(color, ROOK, pindex);
            square[from] = victim; // empty
            pieces[color][pindex] = dest;  // change address
            set_all(affected & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                clear_all(affected & mask[color]);
                square[from].put(color, PAWN, pindex);
                square[dest] = victim; // empty
                pieces[color][pindex] = from; // return to original address
                set_all(affected);
                return 0;
            }
            set_all(affected & mask[other]);
            gamestack.push(move, victim, affected);
            fifty = 0;
            enpassant = 0;
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[color][ROOK][dest];
            material[color] += ROOK_VAL-PAWN_VAL;
            force[color] += 5;
            break;
//----------------------------------------------------
        case PROMO2B:
            affected |= P2BIT(pindex, color) | control[dest];
            clear_all(affected);
            square[dest].put(color, BISHOP, pindex);
            square[from] = victim; // empty
            pieces[color][pindex] = dest;  // change address
            set_all(affected & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                clear_all(affected & mask[color]);
                square[from].put(color, PAWN, pindex);
                square[dest] = victim; // empty
                pieces[color][pindex] = from; // return to original address
                set_all(affected);
                return 0;
            }
            set_all(affected & mask[other]);
            gamestack.push(move, victim, affected);
            fifty = 0;
            enpassant = 0;
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[color][BISHOP][dest];
            material[color] += BISHOP_VAL-PAWN_VAL;
            force[color] += 3;
            break;
//---------------------------------------------------------
        case PROMO2N:
            affected |= P2BIT(pindex, color) | control[dest];
            clear_all(affected);
            square[dest].put(color, KNIGHT, pindex);
            square[from] = victim; // empty
            pieces[color][pindex] = dest;  // change address
            set_all(affected & mask[color]);
            if(control[pieces[color][KING]] & mask[color]){ // in check? undo
                clear_all(affected & mask[color]);
                square[from].put(color, PAWN, pindex);
                square[dest] = victim; // empty
                pieces[color][pindex] = from; // return to original address
                set_all(affected);
                return 0;
            }
            set_all(affected & mask[other]);
            gamestack.push(move, victim, affected);
            fifty = 0;
            enpassant = 0;
            key ^= zobrist[color][ptype][from];
            key ^= zobrist[color][KNIGHT][dest];
            material[color] += KNIGHT_VAL-PAWN_VAL;
            force[color] += 3;
            break;
//----------------------------------------------------
        case NOMOVE:
        break;
    }
   
    /*
        move is legal and undo info on stack
    */
    // must 'move' pieces in castle also
    //clraffected(from, color, ptype, P2BIT(pindex, color));
    //setaffected(dest, color, ptype, P2BIT(pindex, color));
            
    side ^= 1;
    xside ^= 1;
    
	//assert(key == hash());

	key ^= zcastle[castleo & 15];
    key ^= zcastle[castle & 15];
    key ^= zenpassant[enpassanto];
    key ^= zenpassant[enpassant];

    key = ~key;
    
	assert(key == hash());

	ply++;
    
    //assert(victim.ptype != KING);
    //if(key != board.hash()) cout << "\a\a";
	/*if(!isok()){
		display();
		puts("after domove:\n");
		cout << from << " " << dest << " " << color << " " << move->mtype << endl;
		cout << material[WHITE] << " " << material[BLACK] << endl;
		unmove();
		display();
		getchar();
	}*/
    
    return 1;
}
