#include "chess.h"
#include "data.h"
#include "weights.h"
//#include <math.h>

/*
    King Safety sub-terms
        Shield
        MAX(proximity, location)
        hits [number pieces attacking kings's 9'
		pawn 'storm'
        king has limited mobility {2d, linear only, none}
        
		no shield = 54-60
		MAX( center of board = 32+ ,	proximity = 16)
		R+M attack = 3+2 = 5
		pawn attacker = 10

		107/8 = 12
		107/4 = 25
		
*/
// q=12 r=5 m=3
//*

//#define STATICMATE // for null move when dropping into csearch()
const int scale[128] ={
//                        R  2m   *  rm  3m  2r r2m q/4m 2rm r3m qm
//	  0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15

	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  6,  6, 10, 10, 10, 14,
	 16, 18, 20, 22, 24, 25, 26, 27, 28, 29, 29, 29, 30, 30, 30, 31,

//	 16  17  18  19  20  21  22  23  24  25  26  27  28  29  30  31
//	2R2m QR Q2m 2R3m QRm Q3mQ2R -2mr-2r -3m -rm   * -2m  -r   *  -m
	 31, 31, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32};
//*/
/*
const int scale[128] ={//
	  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
	 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32};
*/

void cBOARD::eval_king_safety(){
    int sq;
    int r, f, s, pa, i, d;

    unsigned int attack;
    
    sq = pieces[WHITE][KING];

#ifdef STATICMATE
	if((d=border[sq])<8 && piece_count[BLACK][QUEEN] > 0){
		if(attack = (control[sq+delta[d]] & 0xffff0000)){
			if(attack & attack_mask[BLACK][QUEEN]){
				if(attack ^ attack_mask[BLACK][QUEEN]){
					if(control[sq+delta[d]] & 0x0000fffe);
					else{
						if(side == BLACK){
							liability[WHITE] += 1000;// certainly bad - mated in one
							return;
						}
						liability[WHITE] += 30;// a tempo at least
					}
				}
			}
		}
		if(d < 4){ // corners include 2 more squares
			if(attack = (control[sq+delta[d+4]] & 0xffff0000)){
				if(attack & attack_mask[BLACK][QUEEN]){
					if(attack ^ attack_mask[BLACK][QUEEN]){
						if(control[sq+delta[d+4]] & 0x0000fffe);
						else{
							if(side == BLACK){
								liability[WHITE] += 1000;// certainly bad - mated in one
								return;
							}
							liability[WHITE] += 30;// a tempo at least
						}
					}
				}
			}
			if(attack = (control[sq+delta[((d+1)&3)+4]] & 0xffff0000)){
				if(attack & attack_mask[BLACK][QUEEN]){
					if(attack ^ attack_mask[BLACK][QUEEN]){
						if(control[sq+delta[((d+1)&3)+4]] & 0x0000fffe);
						else{
							if(side == BLACK){
								liability[WHITE] += 1000;// certainly bad - mated in one
								return;
							}
							liability[WHITE] += 30;// a tempo at least
						}
					}
				}
			}
		}
	}
#endif
	
	//cout << "WHITE KING" << endl << "locat\tfreedom\tstorm\tshield\tproxim\thits\ttotal\tcurve\n";
	 
	// do these before we 'adjust' king's location
	s = king_defects[WHITE][sq];

    //cout << s << "\t";

    attack = 0;
    i = 0;
	r = 1;
    do{
        d = sq + delta[i];
        if((d & OFFBOARD) || (control[d] & 0xffff0000) > 0 || square[d].color == WHITE);
		else attack |= r;
		if(i == 7) break;
        i++;
		r <<= 1;
    }while(true);
	
	//cout << attack << "\t";

    pa = 0;
    r = sq >> 4;
    f = (sq & 7) + 1;
    if(f==1){ f = 2; sq++; } // pro: doesnt move into corner
    if(f==8){ f = 7; sq--; } // con: 
    
    f--;
    if(least_adv_pawn[WHITE][f] == 0){  // white pawn is gone 
        //la += lat_onfile[BLACK][f] * ropenbyking; // open file
        if(least_adv_pawn[BLACK][f] == 7) s += shield5; // black pawn is gone too
        else s += shield4;
    }else{ // white pawn is here somewhere...
        //la += lat_onfile[BLACK][f] * rhopenbyking; // half open file
		if(r > 3){
			if(square[sq-17].iswhitepawn());// no penalty
			else if(square[sq-1].iswhitepawn()) s += shield0;
			else if(square[sq-33].iswhitepawn()) s += shield1;
			else if(square[sq-49].iswhitepawn()) s += shield2;
			else s += shield3;
		}else s += shield4;
    }
    if(most_adv_pawn[BLACK][f] > 0 && most_adv_pawn[BLACK][f] <= r) // avoid negative values
        pa += pawn_aggressor[r-most_adv_pawn[BLACK][f]];
    
    f++;
    if(least_adv_pawn[WHITE][f] == 0){
        //la += lat_onfile[BLACK][f] * ropenbyking; // open file
        if(least_adv_pawn[BLACK][f] == 7) s += shield6;
        else s += shield5;
    }else{
        //la += lat_onfile[BLACK][f] * rhopenbyking; // half open file
        if(r > 3){
			if(square[sq-16].iswhitepawn());// no penalty
			else if(square[sq-0].iswhitepawn()) s += shield0;
			else if(square[sq-32].iswhitepawn()) s += shield1;
			else if(square[sq-48].iswhitepawn()) s += shield2;
			else s += shield3;
		}else s += shield5;
    }
    if(most_adv_pawn[BLACK][f] > 0 && most_adv_pawn[BLACK][f] <= r) // avoid negative values
        pa += pawn_aggressor[r-most_adv_pawn[BLACK][f]];
    
    f++;
    if(least_adv_pawn[WHITE][f] == 0){
        //la += lat_onfile[BLACK][f] * ropenbyking; // open file
        if(least_adv_pawn[BLACK][f] == 7) s += shield5;
        else s += shield4;
    }else{
        //la += lat_onfile[BLACK][f] * rhopenbyking; // half open file
        if(r > 3){
			if(square[sq-15].iswhitepawn());// no penalty
			else if(square[sq+1].iswhitepawn()) s += shield0;
			else if(square[sq-31].iswhitepawn()) s += shield1;
			else if(square[sq-47].iswhitepawn()) s += shield2;
			else s += shield3;
		}else s += shield4;
    }
    if(most_adv_pawn[BLACK][f] > 0 && most_adv_pawn[BLACK][f] <= r) // avoid negative values
        pa += pawn_aggressor[r-most_adv_pawn[BLACK][f]];
    
	//cout << pa << "\t" << s << "\t";
	s += pa;
	//s = s * force[BLACK] / MAXFORCE;
	s = s * scale[force[BLACK]] / 32;

	if(attack){
		if(attack & (attack-1)){
			switch(attack){
			case 5:
			case 10:
			case 80:
			case 160: 
				s += no_escape >> 1;
			//	break;
			//default: attack = 0;
			}
		}else s += no_escape >> 1;
	}else s += no_escape;
    
	// q=16 p=8 q/2p + help += 16
	d = proximity[BLACK]; // prox of BLACK pieces to WHITE king
	d += (int)(d>19) << 3; // 8 point bonus 

	//cout << d << "\t";

	d -= defenders[WHITE];
	if(d > 0) s += d;

	// Q=4 R=4 m=2
    //1m  1M       1ea       2M
	//2=2 4=8 5=12 6=18 7=24 8=32 9=40 10=50 12=72  <--- old
	//2=2 4=8 5=10 6=20 7=22 8=24 9=26 10=28 12=32
	//2=2 4=8 5=10 6=12 7=14 8=24 9=26 10=28 12=32
	d = hits[BLACK]; // BLACK piece attacks on WHITE king space
	d += (int)(d>7) << 2;
	s += d;// << 1; 
	//s += ((d * d) >> 2);

	//cout << d << "\t" << s << "\t";

	s = king_safe[s];

	//cout << s << endl;
    
	liability[WHITE] += s;

    
    
    sq = pieces[BLACK][KING];

#ifdef STATICMATE
	if((d=border[sq])<8 && piece_count[WHITE][QUEEN] > 0){
		if(attack = (control[sq+delta[d]] & 0x0000ffff)){
			if(attack & attack_mask[WHITE][QUEEN]){
				if(attack ^ attack_mask[WHITE][QUEEN]){
					if(control[sq+delta[d]] & 0xfffe0000);
					else {
						if(side == WHITE){
							liability[BLACK] += 1000; 
							return;
						}
						liability[BLACK] += 30;
					}
				}
			}
		}
		if(d < 4){ // corners include 2 more squares
			if(attack = (control[sq+delta[d+4]] & 0x0000ffff)){
				if(attack & attack_mask[WHITE][QUEEN]){
					if(attack ^ attack_mask[WHITE][QUEEN]){
						if(control[sq+delta[d+4]] & 0xfffe0000);
						else{
							if(side == WHITE){
								liability[BLACK] += 1000;// certainly bad - mated in one
								return;
							}
							liability[BLACK] += 30;// a tempo at least
						}
					}
				}
			}
			if(attack = (control[sq+delta[((d+1)&3)+4]] & 0x0000ffff)){
				if(attack & attack_mask[WHITE][QUEEN]){
					if(attack ^ attack_mask[WHITE][QUEEN]){
						if(control[sq+delta[((d+1)&3)+4]] & 0xfffe0000);
						else{
							if(side == WHITE){
								liability[BLACK] += 1000;// certainly bad - mated in one
								return;
							}
							liability[BLACK] += 30;// a tempo at least
						}
					}
				}
			}
		}
	}
#endif

	s = king_defects[BLACK][sq];
    
    attack = 0;
    i = 0;
	r = 1;
    do{
        d = sq + delta[i];
        //attack += ((d & OFFBOARD) || (control[d] & 0x0000ffff) > 0 || square[d].color == BLACK);
        if((d & OFFBOARD) || (control[d] & 0x0000ffff) > 0 || square[d].color == BLACK);
		else attack |= r;
		if(i == 7) break;
        i++;
		r <<= 1;
    }while(true);
    
    
    pa = 0;
    r = sq >> 4;
    f = (sq & 7) + 1;
    if(f==1){ f = 2; sq++; } // pro: doesnt move into corner
    if(f==8){ f = 7; sq--; } // con: 
    
    f--;
    if(least_adv_pawn[BLACK][f] == 7){
        //la += lat_onfile[WHITE][f] * ropenbyking; // open file
        if(least_adv_pawn[WHITE][f] == 0) s += shield5;
        else s += shield4;
    }else{
        //la += lat_onfile[WHITE][f] * rhopenbyking; // half open file
		if(r<4){
			if(square[sq+15].isblackpawn()); // no penalty
			else if(square[sq-1].isblackpawn()) s += shield0;
			else if(square[sq+31].isblackpawn()) s += shield1;
			else if(square[sq+47].isblackpawn()) s += shield2;
			else s += shield3;
		}else s += shield4;
    }
    if(most_adv_pawn[WHITE][f] < 7 && most_adv_pawn[WHITE][f] >= r) // avoid negative values
        pa += pawn_aggressor[most_adv_pawn[WHITE][f]-r];
    
    f++;
    if(least_adv_pawn[BLACK][f] == 7){
        //la += lat_onfile[WHITE][f] * ropenbyking; // open file
        if(least_adv_pawn[WHITE][f] == 0) s += shield6;
        else s += shield5;
    }else{
        //la += lat_onfile[WHITE][f] * rhopenbyking; // half open file
        if(r<4){
			if(square[sq+16].isblackpawn()); // no penalty
			else if(square[sq-0].isblackpawn()) s += shield0;
			else if(square[sq+32].isblackpawn()) s += shield1;
			else if(square[sq+48].isblackpawn()) s += shield2;
			else s += shield3;
		}else s += shield5;
    }
    if(most_adv_pawn[WHITE][f] < 7 && most_adv_pawn[WHITE][f] >= r) // avoid negative values
        pa += pawn_aggressor[most_adv_pawn[WHITE][f]-r];
    
    f++;
    if(least_adv_pawn[BLACK][f] == 7){
        //la += lat_onfile[WHITE][f] * ropenbyking; // open file
        if(least_adv_pawn[WHITE][f] == 0) s += shield5;
        else s += shield4;
    }else{
        //la += lat_onfile[WHITE][f] * rhopenbyking; // half open file
        if(r<4){
			if(square[sq+17].isblackpawn()); // no penalty
			else if(square[sq+1].isblackpawn()) s += shield0;
			else if(square[sq+33].isblackpawn()) s += shield1;
			else if(square[sq+49].isblackpawn()) s += shield2;
			else s += shield3;
		}else s += shield4;
    }
    if(most_adv_pawn[WHITE][f] < 7 && most_adv_pawn[WHITE][f] >= r) // avoid negative values
        pa += pawn_aggressor[most_adv_pawn[WHITE][f]-r];
    
    s += pa;
	
	//s = s * force[WHITE] / MAXFORCE;
	s = s * scale[force[WHITE]] / 32;

	if(attack){
		if(attack & (attack-1)){
			switch(attack){
			case 5:
			case 10:
			case 80:
			case 160: 
				s += no_escape >> 1;
			//	break;
			//default: attack = 0;
			}
		}else s += no_escape >> 1;
	}else s += no_escape;
	
    d = proximity[WHITE];
	d += (int)(d>19) << 3; // << piece_count[WHITE][QUEEN]

	d -= defenders[BLACK];
	if(d > 0) s += d;

	// Q=4 R=4 m=2
    //1m  1M       1ea       2M
	//2=2 4=8 5=12 6=18 7=24 8=32 9=40 10=50 12=72  <--- old
	//2=2 4=8 5=10 6=20 7=22 8=24 9=26 10=28 12=32
	//2=2 4=8 5=10 6=12 7=14 8=24 9=26 10=28 12=32
	d = hits[WHITE];
	d += (int)(d>7) << 2;
	s += d;// << 1; 
	//s += ((d * d) >> 2);
	//cout << (d << 1) << endl;	
	
	/*
	if(s > qdepth){
		qdepth = s;
		display();
		cout << s << endl;
		getchar();
	}//*/

	s = king_safe[s];
	liability[BLACK] += s;
	
}
