#include "chess.h"
#include "data.h"
#include "weights.h"

/*
    absolute values:
        CM, stalemate, draw
    decisive values:
        winning exchange, winning a piece, promotion
    non-linear speculative values:
        winning pawn, initiative, king safety issues, passed pawn
    theoretical positional values:
        limited mobility, piece-square
        
    do larger 'profits' lie beyond?
    often, gain is found near larger _threats_ due to limitations imposed by them.
    ie, where a serious loss _looms_, smaller gains are made
    AND where smaller advantages add up, larger threats are found
    
    
    note: static eval is never called when in check.

	Now, features are divided into advantage and disadvantages
	in the hope that certain things may be more synergistic
*/

int cBOARD::evaluate(){
    //return material[side]-material[xside];
	assert(isok()==0);
	
	//if(incheck()) cout << "!";
//	assert(!incheck());
	
	int s, m1, m2, md, ms, t1, t2, t3, t4, i, sq;

    m1 = material[side];
	m2 = material[xside];

	ms = m1+m2; // material sum
	md = m1-m2; // material difference

	if(ms < 651){
		if(ms == 0) return 0; // lone kings
		if(ms == PAWN_VAL) return find_pawn(); // works mostly but sure there are bugs and holes
		if(ms == 325) return 0; // k[b/n] vs k
		if(ms == 650 && md == 0) return 0; // k b/n  vs k b/n 
	}
			
	
	
	/*
		LAZY eval is replaced by 'futility' pruning in isearch() and csearch()
		where we know whether iether side is in check - plus save the call to csearch()
	*/

	
	//beyond this point we need to know piececounts for each type
	// and MOST of the time, we'll need the attack masks too [for pieces] so doit

	piece_count[WHITE][PAWN] = 0;
    piece_count[WHITE][KNIGHT] = 0;
    piece_count[WHITE][BISHOP] = 0;
    piece_count[WHITE][ROOK] = 0;
    piece_count[WHITE][QUEEN] = 0;
    
    piece_count[BLACK][PAWN] = 0;
    piece_count[BLACK][KNIGHT] = 0;
    piece_count[BLACK][BISHOP] = 0;
    piece_count[BLACK][ROOK] = 0;
    piece_count[BLACK][QUEEN] = 0;

    attack_mask[WHITE][PAWN] = 0;
    attack_mask[WHITE][KNIGHT] = 0;
    attack_mask[WHITE][BISHOP] = 0;
    attack_mask[WHITE][ROOK] = 0;
    attack_mask[WHITE][QUEEN] = 0;
    
    attack_mask[BLACK][PAWN] = 0;
    attack_mask[BLACK][KNIGHT] = 0;
    attack_mask[BLACK][BISHOP] = 0;
    attack_mask[BLACK][ROOK] = 0;
    attack_mask[BLACK][QUEEN] = 0;

	
    
	i = 1;
	t2 = 0x00000002;
    do{
        if((sq = pieces[WHITE][i]) & 0x88);
        else{
            t1 = square[sq].ptype;
            attack_mask[WHITE][t1] |= t2;//(1 << i);
            piece_count[WHITE][t1]++;
        }
        if(i == 15) break;
        i++;
		t2 <<= 1;
    }while(true);
    
	i = 1;
	t2 = 0x00020000;
    do{
        if((sq = pieces[BLACK][i]) & 0x88);
        else{
            t1 = square[sq].ptype;
            attack_mask[BLACK][t1] |= t2;//(1 << i);
            piece_count[BLACK][t1]++;
        }
        if(i == 15) break;
        i++;
		t2 <<= 1;
    }while(true);

	// 2minors [not bishop pair] vs k == draw
	if(ms == 650 && (piece_count[side][KNIGHT] > 0 || piece_count[xside][KNIGHT] > 0))
		return 0;


	minors[WHITE] =	piece_count[WHITE][KNIGHT] + piece_count[WHITE][BISHOP];
	minors[BLACK] =	piece_count[BLACK][KNIGHT] + piece_count[BLACK][BISHOP];
	majors[WHITE] =	piece_count[WHITE][ROOK] + piece_count[WHITE][QUEEN];
	majors[BLACK] =	piece_count[BLACK][ROOK] + piece_count[BLACK][QUEEN];

	

#ifdef DEBUGe
	cout << "material:white = " << material[WHITE] << endl
		<< "material:black = " << material[BLACK] << endl;
#endif
	
	
	advantage[WHITE] = 0;
    advantage[BLACK] = 0;
    liability[WHITE] = 0;
    liability[BLACK] = 0;
    
    
    i = 0;
    do{
        least_adv_pawn[WHITE][i] = 0;
        least_adv_pawn[BLACK][i] = 7;
        most_adv_pawn[WHITE][i] = 7;
        most_adv_pawn[BLACK][i] = 0;
        //lat_onfile[WHITE][i] = 0;
        //lat_onfile[BLACK][i] = 0;
		if(i == 9) break;
        i++;
    }while(1);
    
    wpasser = 0;
	bpasser = 0;
    
    //int leftmost, rightmost;
    if(piece_count[WHITE][PAWN]+piece_count[BLACK][PAWN]){
        pre_eval();
		// check for blocked && escorted passed pawns && rooks behind
		//leftmost=8;
		//rightmost=0;
		t1 = 0;
		t4 = advantage[WHITE];// passed pawn score
		
		do{
			if(wpasser&(128>>t1)){
				t2 = most_adv_pawn[WHITE][t1+1];
				//rightmost = t1; // remember last
				//if(t1 < leftmost) leftmost = t1; // latch onto 1st
				t3 = (t2 << 4) + t1;
				t4 += (int)((control[t3] & 0x00000001)>0) * passed_supported[7-t2];
				t4 -= (int)((control[t3] & 0x00010000)>0) * passed_supported[7-t2];
				t4 -= passed_blocked[7-t2] * (square[t3-16].color == BLACK);
				t4 += (int)((control[t3]&control[t3+16]&attack_mask[WHITE][ROOK])>0) * rook_behind_passed;
				t4 -= (int)((control[t3]&control[t3+16]&attack_mask[BLACK][ROOK])>0) * rook_behind_passed/2;
				t4 += (int)(manhattan[pieces[BLACK][KING]+128-t1]>t2-1+side) * 30;
				//cout << t3 << " " << t4 << endl;
			}
			if(t1 == 7) break;
			t1++;
		}while(true);
		
		//t4 += t4 * ((rightmost-leftmost)>1) / (force[BLACK]+1);
		//cout << t4 << endl;
		t4 -= t4 * force[BLACK] / (MAXFORCE<<2); // reduce by 25% on full board
		advantage[WHITE] = t4; 
		
		//leftmost=8;
		//rightmost=0;
		t1 = 0;
		t4 = advantage[BLACK];// passed pawn score
		
		do{
			if(bpasser&(128>>t1)){
				t2 = most_adv_pawn[BLACK][t1+1];
				//rightmost = t1;
				//if(t1 < leftmost) leftmost = t1;
				t3 = (t2 << 4) + t1;
				t4 += (int)((control[t3] & 0x00010000)>0) * passed_supported[t2];
				t4 -= (int)((control[t3] & 0x00000001)>0) * passed_supported[t2];
				t4 -= passed_blocked[t2] * (square[t3+16].color == WHITE);
				t4 += (int)((control[t3]&control[t3-16]&attack_mask[BLACK][ROOK])>0) * rook_behind_passed;
				t4 -= (int)((control[t3]&control[t3-16]&attack_mask[WHITE][ROOK])>0) * rook_behind_passed/2;
				t4 += (int)(manhattan[pieces[WHITE][KING]+16-t1]>7-t2-side) * 30;
				//cout << t3 << " " << t4 << endl;
			}
			if(t1 == 7) break;
			t1++;
		}while(true);
		
		//t4 += t4 * ((rightmost-leftmost)>1) / (force[WHITE]+1);
		//cout << t4 << endl;
		t4 -= t4 * force[WHITE] / (MAXFORCE<<2); // reduce by 25% on full board
		advantage[BLACK] = t4; 
	}
	else{ // no pawns NONE.
		if(force[side]+force[xside] < 17){ // 5+3 vs. 5+3  NOT 5+5 v 3+3 ??
			t1 = force[side]-force[xside];
			if(t1 > -4 && t1 < 4) 
				return(t1*50 + kings2war*(pcsq[KING][side][pieces[side][KING]] - pcsq[KING][xside][pieces[xside][KING]]));
		}
	}
    
#ifdef DEBUGe
	cout << "passed:white = " << advantage[WHITE] << endl
		<< "passed:black = " << advantage[BLACK] << endl;

	cout << "pawn_defect:white = " << liability[WHITE] << endl
		<< "pawn_defect:black = " << liability[BLACK] << endl;
#endif
    	
	// if there ARE any pieces....

	if(force[side]|force[xside]){

		smallest[WHITE] = 0;
		smallest[BLACK] = 0;
		largest[WHITE] = 0;
		largest[BLACK] = 0;

		undeveloped[WHITE] = 0;
		undeveloped[BLACK] = 0;

		proximity[WHITE] = 0;
		proximity[BLACK] = 0;

		defenders[WHITE] = 0;
		defenders[BLACK] = 0;

		minor_attack[WHITE] = attack_mask[WHITE][BISHOP] | attack_mask[WHITE][KNIGHT];
		minor_attack[BLACK] = attack_mask[BLACK][BISHOP] | attack_mask[BLACK][KNIGHT];
		major_attack[WHITE] = attack_mask[WHITE][ROOK] | attack_mask[WHITE][QUEEN];
		major_attack[BLACK] = attack_mask[BLACK][ROOK] | attack_mask[BLACK][QUEEN];

		//											975	 964  903
		//lone pieces and pairs diminished eg 1m also,  3m > 2R > 1Q
		t1 = minors[side] + majors[side];
		if(t1 > 0 && t1 < 3) m1 -= 36 * (3 - t1); 

		t1 = minors[xside] + majors[xside];
		if(t1 > 0 && t1 < 3) m2 -= 36 * (3 - t1); 

		
		// knights worse in endgame - penalize empty board
		t1 = force[side];
		m1 -= piece_count[side][KNIGHT] * (MAXFORCE - t1); 
		
		t2 = force[xside];
		m2 -= piece_count[xside][KNIGHT] * (MAXFORCE - t2); 

		// Pieces given bonus for multitude: eg., minor nearly 4 pawns on full board
		t1 -= 12;//7;
		if(t1 > 0) m1 += t1 * t1; // FULL board += 25^2 = 625  22^2=500

		t2 -= 12;//7;
		if(t2 > 0) m2 += t2 * t2;
		
		//if(m1-m2 > qdepth){ // <= 700
		//	qdepth = m1-m2;
		//	display();
		//	cout << m1-m2;
		//	getchar();
		//}
		
		// bishop pair
		m1 += bishop_pair * (int)(piece_count[side][BISHOP]>1);
		m2 += bishop_pair * (int)(piece_count[xside][BISHOP]>1);
		
		
		//imbalances dependant upon pawn count
		
		t1 = piece_count[side][PAWN];
		t2 = piece_count[xside][PAWN];
		// freedom[] = {18 ... -18}
		
		/*
			The following code handles situations regarding Bishops and 
			Knights: 
			1. knights are better in closed [many pawns?] +-18
			2. bishops better in open [fewer pawns] +-9
			3. the B v K endgame advantage is here 27 cp plus 29 for lonely knight [see above]
			4. the B pair varries from 32 to 68 cp in end  [ie, +-18 cp]

			cut bishop crowd code by 2, since we're doing pawns on sq color code

		*/

		// Bishops improve in open *** handled with pawn of same colors ***
		m1 += freedom[t1] * piece_count[side][BISHOP]/2;
		m2 += freedom[t2] * piece_count[xside][BISHOP]/2;

		
		// Knights improve in crowd
		m1 -= freedom[t1] * piece_count[side][KNIGHT];
		m2 -= freedom[t2] * piece_count[xside][KNIGHT];

		
		// rooks decline in crowd. improve in open board
		// rook on open board will already ALWAYS get ropen bonus
		m1 += freedom[t1] * piece_count[side][ROOK];
		m2 += freedom[t2] * piece_count[xside][ROOK];

		
		//  11+5p vs 12+3p
		s = m1 - m2; // eg., 475 * (3125/100 - 5*2) / 100 = 99
		if(s > force[xside]) s -= force[xside] - t1; //(s * (m2 / 100 - t1)) / 100;
		else if(s < -force[side]) s += force[side] - t2; //(s * (m1 / 100 - t2)) / 100;
		
#ifdef DEBUGe
	cout << "old mdiff = " << md << " new mdiff = " << m1-m2 << " augmented = " << s << endl;
#endif

		eval_pieces();

#ifdef DEBUGe
	cout << "after piece evaluation\n";
	cout << "advantage:white = " << advantage[WHITE] << endl
		<< "advantage:black = " << advantage[BLACK] << endl;

	cout << "liability:white = " << liability[WHITE] << endl
		<< "liability:black = " << liability[BLACK] << endl;
#endif

		//   B        W       B      W
		//128 64    32 16    8 4    2 1
		//if(couldcastle & 3) ((castle & 51) == 0) * lost_castle;
		//((castle & 204) == 0) * lost_castle;

		eval_king_safety();

#ifdef DEBUGe
	cout << "after king safety\n";
	cout << "advantage:white = " << advantage[WHITE] << endl
		<< "advantage:black = " << advantage[BLACK] << endl;

	cout << "liability:white = " << liability[WHITE] << endl
		<< "liability:black = " << liability[BLACK] << endl;
#endif



		
		//king aggression
		// TODO: make k2w_scale[] indexed from force
		t1 = kings2war * pcsq[KING][side][pieces[side][KING]];
		t1 -= t1 * force[xside] / MAXFORCE;
		t2 = kings2war * pcsq[KING][xside][pieces[xside][KING]];
		t2 -= t2 * force[side] / MAXFORCE;
		s += t1 - t2;

		s += count_ctrl() * 3;

	}else{// no pieces
		s = m1 - m2;
		t1 = kings2war * pcsq[KING][side][pieces[side][KING]];
		t2 = kings2war * pcsq[KING][xside][pieces[xside][KING]];
		s += t1 - t2;
	}

	t1 = advantage[side];
	t2 = advantage[xside];
    s += t1 - t2; // pcsq & rhopen & passed pawns
    
	//if(t1-t2 > qdepth){ // <= 300
	//	qdepth = t1-t2;
	//	cout << t1-t2;
	//}

    t1 = liability[side];
    t2 = liability[xside];
    s -= t1 - t2;  // KS, pawn defects
	
	//s += count_ctrl() * 3;

	//if(t1-t2 > qdepth){ // <= 500
	//	qdepth = t1-t2;
	//	cout << t1-t2;
	//}
    
	//*
	ms = force[side] + force[xside];
	md = force[side] - force[xside];

	// if total material on board < 2R+2m or 5m
	if(ms < 17){ // <= Rm v Rm  not QRM v QRM what of overwhelming KS!!
		if(s > 0){// full eval suggests side is winning
			if(md < 4 && piece_count[side][PAWN] == 0) s /= 2;
		}
		else if(s < 0){
			if(md > -4 && piece_count[xside][PAWN] == 0) s /= 2;
		}
	}
	//*/
	
	/*if((t1 = s - material[WHITE]+material[BLACK]) > qdepth){
		qdepth = t1;
		cout << t1;
	}*/
	
	assert(s > -MATE+300 && s < MATE-300);

	return s; 
}    
