/*
    generate pseudo-legal moves storing them in movelist starting at 
    mindex[ply-1] [next available slot to prev ply]
*/
/*
    just as a beginner learns simply where his pieces CAN go, and then, where 
    NOT to put them [ie, en prise] and then where they ought to go [ie, good 
    squares] and then there are pins etc which permit exceptions....
    the point is: MORE OFTEN TRUE THAN FALSE
    remember: search investigates everything, and exceptions are far less 
    frequent than the rules. therefore, more often then not, an attacked piece 
    is lost [EXCEPT!! if it is protected AND the swap() 'adds-up']
    move attacked pieces
    do not move into attacked squares
    but move attacked b before taking undefended p
*/
#include "chess.h"
#include "data.h"
#include "moves.h"
#include "gen.h"


void cMOVELIST::gen(){
    int sq, pc = 0;
    nextmove[ply] = nextmove[ply-1];
    if(side == WHITE){
        do{
            sq = board.pieces[WHITE][pc];
            if((sq & 0x88) == 0){
                switch(board.square[sq].ptype){
                    case KING: genwk(sq);
                    break;
                    case QUEEN: genwq(sq);
                    break;
                    case ROOK: genwr(sq);
                    break;
                    case BISHOP: genwb(sq);
                    break;
                    case KNIGHT: genwn(sq);
                    break;
                    case PAWN: genwp(sq);
                    break;
                    case EMPTY:  
                    cout << "ply:" 
                    << ply << " no white piece["
                    << pc << "] to gen moves for at " 
                    << sq << " ";
                    do{
                        board.display();
                        board.unmove();
                        getchar();
                    }while(ply>1);
                    break;
                }
            }
            if(pc == 15) break;
            pc++;
        }while(1);
    }
    else{
        do{
            sq = board.pieces[BLACK][pc];
            if((sq & 0x88) == 0){
                switch(board.square[sq].ptype){
                    case KING: genbk(sq);
                    break;
                    case QUEEN: genbq(sq);
                    break;
                    case ROOK: genbr(sq);
                    break;
                    case BISHOP: genbb(sq);
                    break;
                    case KNIGHT: genbn(sq);
                    break;
                    case PAWN: genbp(sq);
                    break;
                    case EMPTY: 
                    cout << "ply:" << ply 
                    <<" no black piece to gen moves for at " << sq << " ";
                    //board.display();
                    //cMOVE m = board.gamestack.top();
                    //m.print();
                    getchar(); 
                    break;
                }
            }
            if(pc == 15) break;
            pc++;
        }while(1);
    }
}

void cMOVELIST::addmove(int from, int dest, int mtype, int score){
    if(hashmove[ply].to == dest && hashmove[ply].from == from && hashmove[ply].mtype == mtype)
        score = 1000000;
	//else if(pv[ply][ply+2].to == dest && pv[ply][ply+2].from == from && pv[ply][ply+2].mtype == mtype)
	//    score = 999999;
	else if(killer1[ply].to == dest && killer1[ply].from == from && killer1[ply].mtype == mtype)
        score = MAXHISTSCORE + 30; //BASECAPSCORE-1;
	else if(killer2[ply].to == dest && killer2[ply].from == from && killer2[ply].mtype == mtype)
        score = MAXHISTSCORE + 20; //BASECAPSCORE-2;
	else{
		int t = board.square[from].ptype;
		score += (pcsq[t][side][dest] - pcsq[t][side][from]);
		score += historic[side][from][dest] >> maxhist[side];
	}
    

    nextmove[ply]->set(from, dest, mtype, score);
    nextmove[ply]++;
    //cout << nextmove[ply]-nextmove[0] << " ";
#ifdef debug
    if((from & 0x88) || (dest & 0x88)){
        cout << "generating movement off board \n";
        board.display();
        cout << ply << " from " << from << " dest " << dest << " mtype " << mtype << endl;
        getchar();
    }
#endif
}

void genwk(int from){
    int dest, i = 0;
    do{
        dest = from + delta[i];
        if((dest & OFFBOARD) == 0 && (board.control[dest] & 0xffff0000) == 0){
            if(board.square[dest].color == BLACK)
                tree.addmove(from, dest, CAPTURE, BASECAPSCORE+capval[board.square[dest].ptype]); 
            else if(board.square[dest].color == NONE)
                tree.addmove(from, dest, NORMAL, -4);
        }
        if(i == 7) break;
        i++;
    }while(1);
    if(castle & 1)
        if((board.control[116] & 0xffff0000) == 0)
            if(board.square[117].isempty())
                if((board.control[117] & 0xffff0000) == 0)
                    if(board.square[118].isempty())          
                        if((board.control[118] & 0xffff0000) == 0)
                            tree.addmove(from, 118, CASTLEK, CASTLES);
    if(castle & 2)
        if((board.control[116] & 0xffff0000) == 0)
            if(board.square[115].isempty())
                if((board.control[115] & 0xffff0000) == 0)
                    if(board.square[114].isempty())          
                        if((board.control[114] & 0xffff0000) == 0)
                            if(board.square[113].isempty())
                                tree.addmove(from, 114, CASTLEQ, CASTLES);
}

void genbk(int from){
    int dest, i = 0;
    do{
        dest = from + delta[i];
        if((dest & OFFBOARD) == 0 && (board.control[dest] & 0x0000ffff) == 0){
            if(board.square[dest].color == WHITE)
                tree.addmove(from, dest, CAPTURE, BASECAPSCORE+capval[board.square[dest].ptype]); 
            else if(board.square[dest].color == NONE)
                tree.addmove(from, dest, NORMAL, -4);
        }
        if(i == 7) break;
        i++;
    }while(1);
    if(castle & 4)
        if((board.control[4] & 0x0000ffff) == 0)
            if(board.square[5].isempty())
                if((board.control[5] & 0x0000ffff) == 0)
                    if(board.square[6].isempty())          
                        if((board.control[6] & 0x0000ffff) == 0)
                            tree.addmove(from, 6, CASTLEK, CASTLES);
    if(castle & 8)
        if((board.control[4] & 0x0000ffff) == 0)
            if(board.square[3].isempty())
                if((board.control[3] & 0x0000ffff) == 0)
                    if(board.square[2].isempty())          
                        if((board.control[2] & 0x0000ffff) == 0)
                            if(board.square[1].isempty())
                                tree.addmove(from, 2, CASTLEQ, CASTLES);
}


//-975------------------------------------------------------0----------------------------+975

void genwq(int from){
    int dest, bs = 0, s = 0, i = 0, dlta;
    if(board.control[from] & 0xffff0000)
        bs = FLIGHTQ;//capval[QUEEN]; // RULE: move it or lose it
    do{
        dlta = delta[i];
        dest = from + dlta;
        while((dest & OFFBOARD)==0){
            if(board.square[dest].color == WHITE) break;
            s = bs; // default to history score
            if(board.control[dest] & 0xffff0000) s -= FRIGHTQ; // RULE: we lose piece
            if(board.square[dest].color == BLACK){  // captures
                s += capval[board.square[dest].ptype] + BASECAPSCORE; // RULE: we win
                
				tree.addmove(from, dest, CAPTURE, s);
                break;
            }
            tree.addmove(from, dest, NORMAL, s);
            dest += dlta;
        }
        if(i == 7) break;
        i++;
    }while(1);
}

void genbq(int from){
    int dest, bs = 0, s, i = 0, dlta;
    if(board.control[from] & 0x0000ffff)
     bs = FLIGHTQ;//capval[QUEEN]; // RULE: move it or lose it
    do{
        dlta = delta[i];
        dest = from + dlta;
        while((dest & OFFBOARD)==0){
            if(board.square[dest].color == BLACK) break;
            s = bs; // default to history score
            if(board.control[dest] & 0x0000ffff) s -= FRIGHTQ; /// 10 
            if(board.square[dest].color == WHITE){  // captures
                s += capval[board.square[dest].ptype] + BASECAPSCORE;
                
                tree.addmove(from, dest, CAPTURE, s);
                break;
            }
            tree.addmove(from, dest, NORMAL, s);
            dest += dlta;
        }
        if(i == 7) break;
        i++;
    }while(1);
}

void genwr(int from){
    int dest, bs = 1, s, i = 4, dlta;
    if(board.control[from] & 0xffff0000)
     bs = FLIGHTR; // RULE: move it or lose it
    do{
        dlta = delta[i];
        dest = from + dlta;
        while((dest & OFFBOARD)==0){
            if(board.square[dest].color == WHITE) break;
            s = bs; // reset score of each move to base score
            if(board.control[dest] & 0xffff0000) s -= FRIGHTR; /// 10
            if(board.square[dest].color == BLACK){  // captures
                s += capval[board.square[dest].ptype] + BASECAPSCORE;
                
                tree.addmove(from, dest, CAPTURE, s);
                break;
            }
            tree.addmove(from, dest, NORMAL, s);
            dest += dlta;
        }
        if(i == 7) break;
        i++;
    }while(1);
}

void genbr(int from){
    int dest, bs = 1, s, i = 4, dlta;
    if(board.control[from] & 0x0000ffff)
     bs = FLIGHTR; // RULE: move it or lose it
    do{
        dlta = delta[i];
        dest = from + dlta;
        while((dest & OFFBOARD)==0){
            if(board.square[dest].color == BLACK) break;
            s = bs; // reset score of each move to base score
            if(board.control[dest] & 0x0000ffff) s -= FRIGHTR; /// 10 
            if(board.square[dest].color == WHITE){  // captures
                s += capval[board.square[dest].ptype] + BASECAPSCORE;
                
                tree.addmove(from, dest, CAPTURE, s);
                break;
            }
            tree.addmove(from, dest, NORMAL, s);
            dest += dlta;
        }
        if(i == 7) break;
        i++;
    }while(1);
}

void genwb(int from){
    int dest, bs = 2, s, i = 0, dlta;
    if(board.control[from] & 0xffff0000) bs = FLIGHTM;//capval[BISHOP]; 
    do{
        dlta = delta[i];
        dest = from + dlta;
        while((dest & OFFBOARD)==0){
            if(board.square[dest].color == WHITE) break;
            s = bs; // reset score of each move to base score
            if(board.control[dest] & 0xffff0000) s -= FRIGHTM; /// 10 
            if(board.square[dest].color == BLACK){  // captures
                s += capval[board.square[dest].ptype] + BASECAPSCORE;
                
                tree.addmove(from, dest, CAPTURE, s);
                break;
            }
            tree.addmove(from, dest, NORMAL, s);
            dest += dlta;
        }
        if(i == 3) break;
        i++;
    }while(1);
}

void genbb(int from){
    int dest, bs = 2, s, i = 0, dlta;
    if(board.control[from] & 0x0000ffff) bs = FLIGHTM;
    do{
        dlta = delta[i];
        dest = from + dlta;
        while((dest & OFFBOARD)==0){
            if(board.square[dest].color == BLACK) break;
            s = bs; // reset score of each move to base score
            if(board.control[dest] & 0x0000ffff) s -= FRIGHTM; /// 10 
            if(board.square[dest].color == WHITE){  // captures
                s += capval[board.square[dest].ptype] + BASECAPSCORE;
                
                tree.addmove(from, dest, CAPTURE, s);
                break;
            }
            tree.addmove(from, dest, NORMAL, s);
            dest += dlta;
        }
        if(i == 3) break;
        i++;
    }while(1);
}

void genwn(int from){
    int dest, bs = 2, s, i = 8;
    if(board.control[from] & 0xffff0000) bs = FLIGHTM;
    do{
        dest = from + delta[i];
        if(dest & OFFBOARD);
        else{
            s = bs;
			if(board.control[dest] & 0xffff0000) s -= FRIGHTM; /// 10
            if(board.square[dest].color == BLACK){
                 s += capval[board.square[dest].ptype] + BASECAPSCORE;
                  
                 tree.addmove(from, dest, CAPTURE, s);
            }
            else if(board.square[dest].color == NONE)
                tree.addmove(from, dest, NORMAL, s);
        }
        if(i == 15) break;
        i++;
    }while(1);
}

void genbn(int from){
    int dest, bs = 2, s, i = 8;
    if(board.control[from] & 0x0000ffff) bs = FLIGHTM;
    do{
        dest = from + delta[i];
        if(dest & OFFBOARD);
        else{
            s = bs;
			if(board.control[dest] & 0x0000ffff) s -= FRIGHTM; /// 10 
            if(board.square[dest].color == WHITE){
                s += capval[board.square[dest].ptype] + BASECAPSCORE;
                
                tree.addmove(from, dest, CAPTURE, s);
            }
            else if(board.square[dest].color == NONE)
                tree.addmove(from, dest, NORMAL, s);
        }
        if(i == 15) break;
        i++;
    }while(1);
}

void genwp(int from){
    int bs = 3, s;
    if(from < 40) bs = PAWN27SCORE; // bonus for pawn to 7th
    if(board.square[from - 16].isempty()){
        if(from < 24){ // pawn on 7th
            tree.addmove(from, from-16, PROMO2Q, PROMO2QSCORE);//bs + 875);
            tree.addmove(from, from-16, PROMO2R, -1000);
            tree.addmove(from, from-16, PROMO2B, -1000);
            tree.addmove(from, from-16, PROMO2N, -1000);
        }
        else{
            if(from > 95 && board.square[from - 32].isempty())
                tree.addmove(from, from-32, PPUSH2, 4);
            
            tree.addmove(from, from-16, SINGLE, bs);
        }
    }
    if(((from-17)&0x88)==0 && board.square[from-17].color == BLACK){
        s = capval[board.square[from-17].ptype];
        if(board.control[from - 17] & 0xffff0000) 
            s -= FRIGHTP; /// 10
        if(from<24){
            tree.addmove(from, from-17, PROMOXQ, PROMO2QSCORE+s);
            tree.addmove(from, from-17, PROMOXR, -1000);
            tree.addmove(from, from-17, PROMOXB, -1000);
            tree.addmove(from, from-17, PROMOXN, -1000);
        }
        else{
            tree.addmove(from, from-17, CAPTURE, bs+s + BASECAPSCORE);
        }
    }
    if(((from-15)&0x88)==0 && board.square[from-15].color == BLACK){
        s = capval[board.square[from-15].ptype];
        if(board.control[from - 15] & 0xffff0000) 
            s -= FRIGHTP; /// 10
        if(from<24){
            tree.addmove(from, from-15, PROMOXQ, PROMO2QSCORE+s);
            tree.addmove(from, from-15, PROMOXR, -1000);
            tree.addmove(from, from-15, PROMOXB, -1000);
            tree.addmove(from, from-15, PROMOXN, -1000);
        }
        else{
            tree.addmove(from, from-15, CAPTURE, bs+s + BASECAPSCORE);
        }
    }
    if(enpassant)
        if(enpassant == from-17 || enpassant == from-15){
            s = PAWN_VAL + BASECAPSCORE;
            if(board.control[enpassant] & 0xffff0000) s -= FRIGHTP;
            tree.addmove(from, enpassant, ENPASSANT, s);
        }
}   
    
void genbp(int from){
    int bs = 3, s;
    if(from > 79) bs = PAWN27SCORE;
    if(board.square[from + 16].isempty()){
        if(from > 95){ // pawn on 7th
            tree.addmove(from, from+16, PROMO2Q, PROMO2QSCORE);
            tree.addmove(from, from+16, PROMO2R, -1000);
            tree.addmove(from, from+16, PROMO2B, -1000);
            tree.addmove(from, from+16, PROMO2N, -1000);
        }
        else{
            if(from < 24 && board.square[from + 32].isempty())
                tree.addmove(from, from+32, PPUSH2, 4);
            tree.addmove(from, from+16, SINGLE, bs);
        }
    }
    if(((from+17)&0x88)==0 && board.square[from+17].color == WHITE){
        s = capval[board.square[from+17].ptype];
        if(board.control[from + 17] & 0x0000ffff) 
            s -= FRIGHTP; /// 10
        if(from>95){
            tree.addmove(from, from+17, PROMOXQ, PROMO2QSCORE+s);
            tree.addmove(from, from+17, PROMOXR, -1000);
            tree.addmove(from, from+17, PROMOXB, -1000);
            tree.addmove(from, from+17, PROMOXN, -1000);
        }
        else{
            tree.addmove(from, from+17, CAPTURE, bs+s + BASECAPSCORE);
        }
    }
    if(((from+15)&0x88)==0 && board.square[from+15].color == WHITE){
        s = capval[board.square[from+15].ptype];
        if(board.control[from + 15] & 0x0000ffff) 
            s -= FRIGHTP; /// 10
        if(from>95){
            tree.addmove(from, from+15, PROMOXQ, PROMO2QSCORE+s);
            tree.addmove(from, from+15, PROMOXR, -1000);
            tree.addmove(from, from+15, PROMOXB, -1000);
            tree.addmove(from, from+15, PROMOXN, -1000);
        }
        else{
            tree.addmove(from, from+15, CAPTURE, bs+s + BASECAPSCORE);
        }
    }
    if(enpassant)
        if(enpassant == from+17 || enpassant == from+15){
            s = PAWN_VAL + BASECAPSCORE;
            if(board.control[enpassant] & 0x0000ffff) s -= FRIGHTP;
            tree.addmove(from, enpassant, ENPASSANT, s);
        }
}
    
        

/*   A   B   C   D   E   F   G   H
8    0   1   2   3   4   5   6   7
7   16  17  18  19  20  21  22  23 -
6   32  33  34  35  36  37  38  39
5   48  49  50  51  52  53  54  55 *
4   64  65  66  67  68  69  70  71 *
3   80  81  82  83  84  85  86  87
2   96  97  98  99 100 101 102 103 -
1  112 113 114 115 116 117 118 119
*/ 
