/*
    ttable
*/
#include "chess.h"
#include "data.h"
// 4 8 12 16 20 24 28
// 4 + 8 + 16 + 32 = 60 
void cHASHTABLE::inc_age(){
    age += 4;
    if(age > 60) age = 0;
    stored = 0;
    probed = 0;
    matched = 0;
    useful = 0;
    collisions = 0;
}        

void cHASHTABLE::clear(){
    size_t i = 0;
    do{
        table[i].lock = 0;
        table[i].value = 0;
		table[i].depth = 0;
        table[i].data = 0;
        table[i].bm.from = 0x88;
        table[i].bm.to = 0x88;
    }while(++i <= main_hash_size);

    stored = 0;
    probed = 0;
    matched = 0;
    useful = 0;
    collisions = 0;
    age = 4;
}

int cHASHTABLE::count(){
    size_t i = 0, c = 0;
    do{
        c += ((table[i].data & 3) > 0);
    }while(++i <= main_hash_size);
    return c;
}

cHASHENTRY * cHASHTABLE::probe(){
    cHASHENTRY * e;
    //i64 lock = key;
    //lock ^= zcastle[castle & 15];
    //lock ^= zenpassant[enpassant];
    unsigned int entry = (unsigned int)(key & main_hash_size);
//    probed++;
    e = (table + entry);
    if(e->lock == key){
//        matched++;
        return e;
    }
    return NULL;
}

void cHASHTABLE::store(int depth, int val, int type, cMOVE *bm){
    //mate score cant get over-written because draft is always sufficient to return after probe
	//which may be why we have trouble with 'shortest mates' sometimes
	// but quies can never overwrite mate scores
    unsigned int entry = (unsigned int)(key & main_hash_size);
    cHASHENTRY *tt = (table+entry);
    
	if (val > MATE-300){
		val += (ply-0); 
		depth = 64 - (MATE-val); 
		//type = FAIL_HIGH_LOWER;
	}
    else if (val < -MATE+300){ 
		val -= (ply-0); 
		depth = 64 - (MATE+val); 
		//type = FAIL_LOW_UPPER;
	}
	    
	if(depth >= tt->depth){ //over-write all shallower entries
        stored += ((tt->data & 60) != age);
        tt->lock = key;
        tt->value = val;
        tt->depth = depth;
        tt->data = type | age;
        tt->bm.from = bm->from;
        tt->bm.to = bm->to;
        tt->bm.type = bm->mtype;
    }
    else if((tt->data & 60) != age){
        stored++; // over-write old search stores, though deeper
        tt->lock = key;
        tt->value = val;
        tt->depth = depth;
        tt->data = type | age;
        tt->bm.from = bm->from;
        tt->bm.to = bm->to;
        tt->bm.type = bm->mtype;
    }
    //else collisions++; // if 'draft' was deeper, why didnt we use it?
	// maybe it was an insufficient bound score
}

//*
void cHASHTABLE::stuffmove(cMOVE * bm){
    //i64 lock = key;
    //lock ^= zcastle[castle & 15];
    //lock ^= zenpassant[enpassant];
    unsigned int entry = (unsigned int)(key & main_hash_size);
    table[entry].value = 0;
    table[entry].depth = 0;
	table[entry].lock = key;
    table[entry].data = EXACT_PV | age;
    table[entry].bm.from = bm->from;
    table[entry].bm.to = bm->to;
    table[entry].bm.type = bm->mtype;
}
//*/
