#include "chess.h"
#include "data.h"
#include "weights.h"
#include <math.h>
static int zone[128]={
     0, 1, 6, 6, 6, 6, 1, 0,    0,0,0,0,0,0,0,0,
     1, 2, 3, 3, 3, 3, 2, 1,    0,0,0,0,0,0,0,0,
     6, 3, 4, 4, 4, 4, 3, 6,    0,0,0,0,0,0,0,0,
     6, 3, 4, 5, 5, 4, 3, 6,    0,0,0,0,0,0,0,0,
     6, 3, 4, 5, 5, 4, 3, 6,    0,0,0,0,0,0,0,0,
     6, 3, 4, 4, 4, 4, 3, 6,    0,0,0,0,0,0,0,0,
     1, 2, 3, 3, 3, 3, 2, 1,    0,0,0,0,0,0,0,0,
     0, 1, 6, 6, 6, 6, 1, 0,    0,0,0,0,0,0,0,0};
     
static int value[6][7] = {
//    0   1   2   3   4   5   6    zone    
    { 0,  1,  3,  3,  4,  4,  2}, // kings use zone 6 to flee corners
    { 0,  1,  2,  3,  4,  5,  1}, // queens
    { 0,  0,  0,  0,  0,  0,  0}, // rooks  <--- skip rooks using custom
    { 2,  4,  6, 10, 11, 12,  4}, // bishops  10, 16, 18, 20, 22, 24, 16
    { 0,  4, 12, 12, 16, 18,  1}, // knights  0,  1, 8, 10, 12, 14, 1
    { 0,  0,  0,  0, 16, 25,  0}, // pawns <--- skip pawns using custom 
};

void clear_historic(){
    for(int x = 0; x < 128; x++){
        if(x & 0x88) continue;
        for(int y = 0; y < 128; y++){
            //if(y & 0x88) continue;
            historic[WHITE][x][y] = 0;
            historic[BLACK][x][y] = 0;
        }
    }
	maxhist[side] = 0;
	maxhist[xside] = 0;
}

void age_historic(){
    for(int x = 0; x < 120; x++){
		if(x & 0x88) x += 8;
		for(int y = 0; y < 120; y++){
			if(y & 0x88) y += 8;
			historic[WHITE][x][y] >>= maxhist[WHITE];
			historic[BLACK][x][y] >>= maxhist[BLACK];
		}
	}
	maxhist[side] = 0;
	maxhist[xside] = 0;
}


void init_data(){
	//0.5	0.001	0.0003	9.00E-07
	//0.5	0.001	0.00027	8.00E-07  drops i=150 to just above rook
	//0.7	0.0001	0.00018	5.00E-07  
	//1		5.00E-03	0.0002	5.60E-07
	//0.7	5.00E-03	0.0004	1.10E-06
	//2.50E-08	8.20E-11
	for(int i = 0; i < 256; i++){
		//float n = (float)i + powf(0.055f * (float)i, 3.0f) - powf(0.0265f * (float)i, 4.0f);
		//float n = (float)i + powf(0.1f * (float)i, 2.0f) - powf(0.03f * (float)i, 3.0f);
		//if(i>102) n = i + 75;
		float n = (float)i + powf(0.1f * (float)i, 2.0f);
		//float n = (float)i + powf(0.05f * (float)i, 3.0f);
		
		//float n = (float)i * 0.2f +powf(0.17f * (float)i, 2.0f) - powf(0.053f * (float)i, 3.0f);
		//if(i>125) n = i+60;
		
		//float n = (float)i * 0.1f +powf(0.15f * (float)i, 2.0f) - powf(0.038f * (float)i, 3.0f);
		king_safe[i] = n + 0.5f;
		//cout << i << ":" << king_safe[i] << "\t";
	}
	
	
	for(int i = 0; i < 256; i++) pseudoe[i] = 0;
	side = 0;
	//for(int s = 0; s < 120; s++){
	board.pieces[0][KING] = 0;
	
	for(int d = 1; d < 120; d++){
		if(d&0x88);
		else{
			pseudoe[128+d] = onking(d);
			pseudoe[128-d] = onking(d); // top l&r
		}
	}
	//}
	
	for(int y = 0; y < 7; y++){
		for(int x = 0; x < 8; x++){
			pseudoe[128+16*y+8+x] = pseudoe[96-16*y+8+x];
			pseudoe[112-16*y+x] = pseudoe[144+16*y+x];
		}
	}

	
	
	//for(int i = 0; i < 128; i++){
	//	if(i & 0x88);
	//	else cout << pseudoe[128+51-i];
	//	if((i&15) == 15) cout << endl;
	//}

	//for(int i = 0; i < 256; i++){
	//	//if(i & 0x88);
	//	cout << pseudoe[i];
	//	if((i&15) == 15) cout << endl;
	//}

	board.least_adv_pawn[WHITE][0] = 0;
    board.least_adv_pawn[WHITE][9] = 0;
    board.least_adv_pawn[BLACK][0] = 7;
    board.least_adv_pawn[BLACK][9] = 7;
    
    board.most_adv_pawn[WHITE][0] = 7;
    board.most_adv_pawn[WHITE][9] = 7;
    board.most_adv_pawn[BLACK][0] = 0;
    board.most_adv_pawn[BLACK][9] = 0;
    
    board.attack_mask[WHITE][KING] = 0x00000001;
    board.attack_mask[BLACK][KING] = 0x00010000;
    board.piece_count[WHITE][KING] = 1;
    board.piece_count[BLACK][KING] = 1;
    
    //board.lat_onfile[WHITE][0] = 0;
    //board.lat_onfile[WHITE][9] = 0;
    //board.lat_onfile[BLACK][0] = 0;
    //board.lat_onfile[WHITE][9] = 0;
    
    //srand(777);
    //znullmove = rand64();
    for(int i = 0; i < 16; i++)
        zcastle[i] = rand64();
    for(int i = 0; i < 128; i++)
        zenpassant[i] = rand64();
        
    for(int x = 0; x < 128; x++)
        for(int y = 0; y < 128; y++){
            historic[WHITE][x][y] = 0;
            historic[BLACK][x][y] = 0;
        }
    
    for(int s = 0; s<120; s++)
        if(s & 0x88);
        else{
            for(int p = 0; p<6; p++){
                zobrist[WHITE][p][s] = rand64();
                zobrist[BLACK][5-p][119-s] = rand64();
            }
        }
        
    
    
    for(int y = 0; y < 8; y++){
        //cout << "\n";
        for(int s, x = 0; x < 8; x++){
            s = (x>y)? x : y;
            manhattan[128 + y*16+x] = s;
            manhattan[128 - (y*16+x)] = s;
            //cout << s << " ";
        }
        for(int x, s, xx = 8; xx < 16; xx++){
            x = 7 - (xx & 7);
            s = (x>y)? x : y;
            s++;
            manhattan[128 + y*16+xx] = s;
            manhattan[128 - (y*16+xx)] = s;
            //cout << s << " ";
        }
    }

	
    
	//*
	for(int p = KING; p <= PAWN; p++){  
        if(p == PAWN || p == ROOK /* || p == KING */ ) continue;// using original pawn table
		for(int s=0; s<120; s++){
            if(s&0x88) continue;
			pcsq[p][WHITE][s]=value[p][zone[s]]+(7 - (s >> 4))/4;
            pcsq[p][BLACK][s]=value[p][zone[s]]+(0 + (s >> 4))/4;
        }
	}
	
	

    for(int i=0; i<128; i++) e_scale[i]=0;
	for(int i = 0; i < 32; i++){
		e_scale[31-i] = e_scale[32-i] + i;
	}
	/*
	for(int i = 0; i < 32; i++){
		cout << e_scale[i] << ", ";
        if((i&15)==15) cout << endl;
	}//*/
	

	for(int p = 0; p < EMPTY; p++){
		//cout << endl << endl;
		for(int s = 0; s < 120; s++){
			if(s&8){
				//cout << endl; 
				s += 8;
			}
			if(pcsq[p][WHITE][s] != pcsq[p][BLACK][112-(s&240) + (s&7)]) 
				cout << p << " " << s << endl;
			//cout << pcsq[p][WHITE][s] << "\t";
		}
	}
	//cout << endl << endl;

    
}
