#include "chess.h"
#include "data.h"
#include "weights.h"
#include "search.h"

#ifdef _LINUX
#include <stdlib.h>
#endif

/*
    Intermediate Search
	
	Includes all captures[not underpromotions], Queening, checks, and check evasions!
	
	will detect mate and mate-in-one positions
	NB. unless the position was stored by csearch in hash !!!

	While this is certainly slower, the search is far more stable

*/

int isearch(int alpha, int beta){
    
	assert(board.isok()==0);
	assert(beta > alpha);
	
#ifdef TRACEi
	logfile << key << "\t\tENTi\t" << alpha << "\t" << beta << "\t" << 0 << "\t" << ply << endl;
#endif

	qnodes++;
	
	pvl[ply] = ply;
    if(ply >= MAXPLY-2) return beta;
    if(ply > qdepth) qdepth = ply;
	
	if(fifty > 3){
        if(board.gamestack.rep()) return 0;
        if(fifty >= move50) return 0;
    }
	
	// time check.  @100knps time check every 1/12 sec
    if((qnodes & 8191) == 8191)
        if(clock() >= st /* || Bioskey() */){
            stopsearch = 1;
            return 0;
        }
    
    cMOVE bm;
    int val;
    
    tree.killer1[ply+1].to = 0x88;
    tree.killer2[ply+1].to = 0x88;
    tree.hashmove[ply].to = 0x88;

#ifdef HASHING

	cHASHENTRY * he = hashtable.probe();
    if(he != NULL){
        tree.hashmove[ply].from = he->bm.from;
        tree.hashmove[ply].to = he->bm.to;
        tree.hashmove[ply].mtype = he->bm.type;
        val = he->value;

        if(ABS(val) > MATE-300){
            if(val > 0) val -= (ply-0);
            else val += (ply-0);
        }
		//CANNOT USE VALUES STORED BY CSEARCH()
		if(he->depth){
			switch(he->data & 3){
			case FAIL_LOW_UPPER: // All searched. val has best score
				if(val <= alpha){
					hashtable.stored += ((he->data & 60) != hashtable.age);
					return val;
				}
				break; 
			case FAIL_HIGH_LOWER: // Few searched. val has best
				if(val >= beta){
					hashtable.stored += ((he->data & 60) != hashtable.age);
					return val;
				}
				break;
			case EXACT_PV:
				if(val > alpha && val < beta)
					extract(&tree.hashmove[ply]);
				hashtable.stored += ((he->data & 60) != hashtable.age);
				return val;
				break;
			}
		}
    }
	
#endif

    bm = tree.hashmove[ply];
    cMOVE *cm;

	bool incheck = board.incheck();	
	int nt = FAIL_LOW_UPPER;
	
	if(incheck) val = -MATE +(ply-1);// + 2;//-NFINITY;//
	else val = board.evaluate();

	if(val > alpha){
		if(val >= beta){
			hashtable.store(0, val, FAIL_HIGH_LOWER, &bm);

#ifdef TRACEi
			logfile << key << "\t\tiSTP\t" << alpha << "\t" << beta << "\t" << val << "\t" << 0 << "\t" << ply << endl;
#endif
			return val;
		}
		alpha = val;
	}
	

	tree.gen();
	
	int legal = 0, stdpat = val, best = -NFINITY;
    
	
    cm = tree.nextmove[ply-1];
    while(cm < tree.nextmove[ply]){
        tree.isortone(cm);
        if(!incheck // not in check -> look at all moves[for checks]
			//else, only what may evade
			|| cm->is_pseudo_evasion()
		){ 
										
#ifdef TRACEi
		cm->to_san(buffer);
		i64 lock = key;
#endif
			
			if(board.domove(cm)){
                legal++;
				
				if(incheck
					|| board.incheck()
					// || board.gamestack.was_p27()
					|| (cm->mtype < NORMAL

#ifdef FUTILITYi
					&& board.material[xside]-board.material[side]+max_pos_value > alpha
#endif
					)){

#ifdef TRACEi
						logfile << lock << "\t" << buffer << "\tCSE\t" << -alpha << "\t" << -beta << "\t" << 0 << "\t" << ply << endl;;
#endif
						val = -csearch(-beta, -alpha);
						//
				}
				
				board.unmove();
                
				if(val > best){
					if(val > alpha){  // best move so far 
                        if(val >= beta){ // refutes previous move 
                            if(cm->score < 150) bump(cm, 1);
                            fh1 += (legal == 1);
                            fhn++;
                            hashtable.store(0, val, FAIL_HIGH_LOWER, cm);
#ifdef TRACEi
							cm->to_san(buffer);
							logfile << key << "\t" << buffer << "\tiCUT\t" << alpha << "\t" << beta << "\t" << val << "\t" << 0 << "\t" << ply << endl;
#endif
                            return val;//beta;
                        }
                        alpha = val;
						//bm = *cm;
                        nt = EXACT_PV;
                        /* update the PV */
                        pv[ply][ply] = *cm; // for IID
                        for (int j = ply + 1; j < pvl[ply + 1]; ++j)
                            pv[ply][j] = pv[ply + 1][j];
                        pvl[ply] = pvl[ply + 1];
                    }
                    bm = *cm;
					best = val;
					if(best > stdpat) stdpat = best;
                }
			}
        }
        cm++;
    }
    
	if(legal==0){
		if(incheck){
			stdpat = -MATE+(ply-1);
			nt = EXACT_PV;
#ifdef TRACEi
			logfile << key << "\t*********** iMATE ***********\n";
#endif
			/*board.display();
			cout << "\nisearch()\n";
			getchar();*/
		}//else if(stdpat <= -MATE+300) cout << "danger Will\n";
		//hashtable.store(0, stdpat, nt, &no_move);
		//return stdpat;
	}//else
		
	//if(bm.score < 150) bump(&bm, 1);
	    
	hashtable.store(0, stdpat, nt, &bm);
    

#ifdef TRACEi
	bm.to_san(buffer);
	if(nt == EXACT_PV)
		logfile << key << "\t" << buffer << "\tiPVN\t" << alpha << "\t" << beta << "\t" << stdpat << "\t" << 0 << "\t" << ply << endl;
	else
		logfile << key << "\t" << buffer << "\tiALL\t" << alpha << "\t" << beta << "\t" << stdpat << "\t" << 0 << "\t" << ply << endl;
#endif

	assert(stdpat > -MATE && stdpat < MATE);

	return stdpat;
}

