#include "chess.h"
#include "data.h"
#include "gen.h"


int cMOVE::is_volitile(){
	return (mtype < NORMAL || (mtype > PPUSH2 && mtype < NOMOVE));
}

//int cMOVE::is_capture(){
//	return (mtype < NORMAL);
//}


// forceval[8] = {0,12,5,3,3,0,0,0} donull +- forceval[ptype]  34
int cBOARD::null_ok(){
	return(force[side] > 4);
}


int cMOVE::is_pseudo_evasion(){
	if(board.square[from].ptype == KING) return 1;
	int y = board.pieces[side][KING];
	if(!pseudoe[128+y-to]) return 0;
	// since double checks are so rare, it seems better to catch them with domove()
	//unsigned int c = board.control[y];
	//if(side == WHITE) c >>= 16;
	//else c &= 0x0000ffff;
	//if(popcount16(c) > 1) return 0; // double attack excludes all else
    if(mtype < PROMO2Q) return 1;
	if(pseudoe[128+y-to] == 2) return 0; // knights
    return 1;
}


void cMOVE::to_string(char * s){
    const char pc[]="KQRBNP?";
    int i = 0;
    if(board.square[from].ptype == 5)
        s[i++] = 'a' + (from & 7);
    else s[i++] = pc[board.square[from].ptype];
    if((mtype<=PROMOXQ || (mtype>=PROMOXR && mtype <= PROMOXN))) s[i++] = 'x';
    else if(board.square[from].ptype == 5) i--;     
    s[i++] = 'a' + (to & 7);
    s[i++] = '1' + 7 - (to / 16);
    switch(mtype){
        case PROMO2Q: s[i++] = 'Q';
        break;
        case PROMO2R: s[i++] = 'R';
        break;
        case PROMO2B: s[i++] = 'B';
        break;
        case PROMO2N: s[i++] = 'N';
        break;
        case PROMOXQ: s[i++] = 'Q';
        break;
        case PROMOXR: s[i++] = 'R';
        break;
        case PROMOXB: s[i++] = 'B';
        break; 
        case PROMOXN: s[i++] = 'N';
        break;
        //default:  s[i++] = ' ';
    }
    s[i] = 0;
}

void cMOVE::print(){
    char buffer[8];
    //to_string(buffer);
	to_san(buffer);
    cout << " ";
    cout << buffer;
}

//  convert string to move eg.  e7e8q or g2g1Q

bool s2move(char* s, cMOVE &m){
    m.from = s[0]-'a'+16*(7-(s[1]-'1'));
    m.to = s[2]-'a'+16*(7-(s[3]-'1'));
	m.score = 0;
    cMOVE * i=tree.nextmove[ply-1];
    while(i<tree.nextmove[ply]){
		if(m.from==i->from){
            if(m.to==i->to){
                if(i->mtype == PROMOXQ || i->mtype == PROMO2Q || (i->mtype >= PROMOXR && i->mtype <= PROMO2N)){
					if((s[4] == 'Q' || s[4] == 'q') && (i->mtype == PROMO2Q || i->mtype == PROMOXQ)){
                        m.mtype = i->mtype;
                        return true;
                    }
                    if((s[4] == 'R' || s[4] == 'r') && (i->mtype == PROMO2R || i->mtype == PROMOXR)){
                        m.mtype = i->mtype;
                        return true;
                    }
                    if((s[4] == 'B' || s[4] == 'b') && (i->mtype == PROMO2B || i->mtype == PROMOXB)){
                        m.mtype = i->mtype;
                        return true;
                    }
                    if((s[4] == 'N' || s[4] == 'n') && (i->mtype == PROMO2N || i->mtype == PROMOXN)){
                        m.mtype = i->mtype;
						cout << s[4];
                        return true;
                    }
					i++;
					continue;
				}
                m.mtype= i->mtype;   
                return true;
            }
        }
        i++;
    }
    return false;
}

bool cMOVE::is_pseudo(){
    if(from & 0x88) return false;
    int p = board.square[from].ptype;
    int c = board.square[from].color;
    if(p == EMPTY || c != side) return false;
    // prepare to generate short move list for one piece
    tree.nextmove[ply] = tree.nextmove[ply-1]; 
    if(c == WHITE){
        switch(p){
            case KING: genwk(from);
            break;
            case QUEEN: genwq(from);
            break;
            case ROOK: genwr(from);
            break;
            case BISHOP: genwb(from);
            break;
            case KNIGHT: genwn(from);
            break;
            case PAWN: genwp(from);
            break;
        }
    }else{ // c==BLACK
        switch(p){
            case KING: genbk(from);
            break;
            case QUEEN: genbq(from);
            break;
            case ROOK: genbr(from);
            break;
            case BISHOP: genbb(from);
            break;
            case KNIGHT: genbn(from);
            break;
            case PAWN: genbp(from);
            break;
        }
    }
    // search the list for matching move
    cMOVE * i = tree.nextmove[ply-1];
    while(i < tree.nextmove[ply]){
        if(i->to == to && i->mtype == mtype) return true;
        ++i;
    }
    return false;
}



// 
// if same piece-type
// if moves to same square
// if pieces on same file insert rank
// if pieces on same rank insert file
// requires current board moves generated
int cMOVE::to_san(char * san){
	if((from & 0x88) || (to & 0x88)){ *san = 0; return 0; }
	cMOVE *cur, *end;
	const char p[8] = "KQRBNP_";
	int pt = board.square[from].ptype;
	if(pt == PAWN){*san = 'a' + (from & 7); *(++san) = '1' + 7 - (to >> 4);}
	else *san = p[pt];
	cur = tree.nextmove[ply-1];
	end = tree.nextmove[ply];
	int f = 0;
	while(cur < end){
		if(is_equal(cur)) f = 1; // move is here!
		// iff non-pawn, same type, and same dest, add disambiguation
		else if(pt != PAWN && board.square[cur->from].ptype == pt && cur->to == to){
			if((from&7) != (cur->from & 7)) *(++san) = 'a' + (from & 7);
			else *(++san) = '1' + 7 - (from >> 4);
		}
		cur++;
	}
	if(!f) return 0;

	if((mtype<=PROMOXQ || (mtype>=PROMOXR && mtype <= PROMOXN))){
		if(pt == PAWN){
			*san = 'x'; // pawn push is simply destination sq 'e4'
			*(++san) = 'a' + (to & 7);
			*(++san) = '1' + 7 - (to >> 4);
		}
		else *(++san) = 'x';
	}
	
	if(pt != PAWN){
		*(++san) = 'a' + (to & 7);
		*(++san) = '1' + 7 - (to >> 4);
	}
	

    switch(mtype){
        case PROMO2Q: *(++san) = '='; *(++san) = 'Q';
        break;
        case PROMO2R: *(++san) = '='; *(++san) = 'R';
        break;
        case PROMO2B: *(++san) = '='; *(++san) = 'B';
        break;
        case PROMO2N: *(++san) = '='; *(++san) = 'N';
        break;
        case PROMOXQ: *(++san) = '='; *(++san) = 'Q';
        break;
        case PROMOXR: *(++san) = '='; *(++san) = 'R';
        break;
        case PROMOXB: *(++san) = '='; *(++san) = 'B';
        break; 
        case PROMOXN: *(++san) = '='; *(++san) = 'N';
        break;
        //default:  *(++san) = ' ';
    }
    *(++san) = 0;

	return 1;
}