#include "chess.h"
#include "data.h"
#include "weights.h"

/*
    pre-eval 
        collect info for evaluation pawn hash
        requires:
            liability[] = 0;
            attack_mask[][] set
            piece_count[][] set
        
*/

void cBOARD::pre_eval(){
    int f, r, pc, sq;
    unsigned int pawnkey = 0;
        
    //shade = WHITE;
    pc = 1;
	do{
        if((sq = pieces[WHITE][pc]) & 0x88);
        else if(square[sq].ptype == PAWN){
            f = (sq&7)+1; // adjusted for 10 files {0 .. 9}
            r = sq >> 4;
			//minusb += pcsq[PAWN][WHITE][sq];
            if(r > least_adv_pawn[WHITE][f]) least_adv_pawn[WHITE][f] = r;
            if(r < most_adv_pawn[WHITE][f]) most_adv_pawn[WHITE][f] = r;
            pawnkey ^= zobrist[WHITE][PAWN][sq];
        }
        if(pc==8) break;
        pc++;
    }while(true);
    
    //shade = BLACK;
    pc = 1;
    do{
        if((sq = pieces[BLACK][pc]) & 0x88);
        else if(square[sq].ptype == PAWN){
            f = (sq&7)+1; // adjusted for 10 files {0 .. 9}
            r = sq >> 4;
            //minusw += pcsq[PAWN][BLACK][sq];
            if(r < least_adv_pawn[BLACK][f]) least_adv_pawn[BLACK][f] = r;
            if(r > most_adv_pawn[BLACK][f]) most_adv_pawn[BLACK][f] = r;
            pawnkey ^= zobrist[BLACK][PAWN][sq];
        }
        if(pc==8) break;
        pc++;
    }while(true);

    if(side == BLACK) pawnkey = ~pawnkey;

#ifdef PHASHING    
    cPHASHENTRY * pentry = phashtable.probe(pawnkey);
    
    if(pentry->lock == pawnkey){
        liability[WHITE] = pentry->wliability;
        liability[BLACK] = pentry->bliability;
		advantage[WHITE] = pentry->wpps;//passed pawn score
		advantage[BLACK] = pentry->bpps;//passed pawn score
		wpasser = pentry->wpassed;
		bpasser = pentry->bpassed;
        return;
    }
#endif
	
	int t, cpp,bonuses,defects;
	int plusw = 0, plusb = 0;
	int minusw = 0, minusb = 0;

    //shade = WHITE;
    pc = 1;
    do{
        if((sq = pieces[WHITE][pc]) & 0x88);
        else if(square[sq].ptype == PAWN){
            bonuses = 0;
			defects = 0;
			minusb += pcsq[PAWN][WHITE][sq];
            f = (sq&7)+1; // adjusted for 10 files {0 .. 9} 
            r = sq >> 4;
            //isolated, doubled, backward, advanced, passed
            t = (least_adv_pawn[WHITE][f-1] == 0) + (least_adv_pawn[WHITE][f+1] == 0);
            if(t == 2) defects += isolated_pawn;// >> dbl;
            else{
                defects += doubled_pawn * (r > most_adv_pawn[WHITE][f]);
				defects += t * end_pawn;
                defects += weak_pawn * ((control[sq] & attack_mask[WHITE][PAWN])==0);
                if(!square[sq-16].isblackpawn() && // not blocked
                    ((control[sq-16] & attack_mask[WHITE][PAWN]) > 0 || // advance supported
                    (control[sq-16] & attack_mask[BLACK][PAWN]) == 0)); // not prevented
                else defects += immobile_pawn; //  
            }
            
            if(r < least_adv_pawn[BLACK][f] && r == most_adv_pawn[WHITE][f]){ // no opposing pawn
                t = (r <= least_adv_pawn[BLACK][f-1]) + (r <= least_adv_pawn[BLACK][f+1]);
                if(t == 2){
                    bonuses += passed_pawn[7-r];
					wpasser |= 128 >> (f-1);
                    cpp = control[sq]|control[sq-16];
					if(r>1) cpp |= control[sq-32];
                    bonuses += passed_connected[7-r] * ((cpp&attack_mask[WHITE][PAWN])>0);
                }else bonuses += t * passed_hidden[7-r];
            }//*/   
        
			minusw += defects;// * (100+bonuses)/100;
			plusw += bonuses;
		}
        if(pc==8) break;
        pc++;
    }while(true);
    
    advantage[WHITE] = plusw;
    
    //shade = BLACK;
    pc = 1;
    do{
        if((sq = pieces[BLACK][pc]) & 0x88);
        else if(square[sq].ptype == PAWN){
            bonuses = 0;
			defects = 0;
			minusw += pcsq[PAWN][BLACK][sq];
            f = (sq&7)+1;
            r = sq >> 4;
            //isolated, doubled, backward, advanced, passed
            t = (least_adv_pawn[BLACK][f-1] == 7) + (least_adv_pawn[BLACK][f+1] == 7);
            if(t == 2) defects += isolated_pawn;// >> dbl;
            else{
                defects += doubled_pawn * (r < most_adv_pawn[BLACK][f]);
				defects += t * end_pawn;
                defects += weak_pawn * ((control[sq] & attack_mask[BLACK][PAWN])==0);
                if(!square[sq+16].iswhitepawn() && // not blocked
                    ((control[sq+16] & attack_mask[BLACK][PAWN]) > 0 || // supported
                    (control[sq+16] & attack_mask[WHITE][PAWN]) == 0)); // not prevented
                else defects += immobile_pawn; // 
            }
            
            if(r > least_adv_pawn[WHITE][f] && r == most_adv_pawn[BLACK][f]){ // no opposing pawn
                t = (r >= least_adv_pawn[WHITE][f-1]) + (r >= least_adv_pawn[WHITE][f+1]);
                if(t == 2){
                    bonuses += passed_pawn[r];
					bpasser |= 128 >> (f-1);
					//cout << f << " " << r << endl;
                    cpp = control[sq]|control[sq+16];
					if(r<6) cpp |= control[sq+32];
                    bonuses += passed_connected[r] * ((cpp&attack_mask[BLACK][PAWN])>0);
                }else bonuses += t * passed_hidden[r];
            }//*/   
        
			minusb += defects;// * (100+bonuses)/100;
			plusb += bonuses;
		}
		if(pc==8) break;
        pc++;
    }while(true);
    
    advantage[BLACK] = plusb;
    
	liability[WHITE] = minusw;
	liability[BLACK] = minusb;

#ifdef PHASHING
	pentry->wliability = minusw;
    pentry->bliability = minusb;
	pentry->wpps = advantage[WHITE];
    pentry->bpps = advantage[BLACK];
	pentry->wpassed = wpasser;
	pentry->bpassed = bpasser;
#endif

    return;
}
