#include "chess.h"
#include "data.h"
#include "weights.h"
#include "search.h"

#ifdef _LINUX
#include <stdlib.h>
#endif


/*
    capture search
	captures, promotions[Q ONLY]
	will NOT detect mate
	returns a minimum of stand-pat static eval unless mated
*/

int qsearch(int alpha, int beta){
    
	assert(board.isok()==0);
	
	qnodes++;

#ifdef TRACEc
	logfile << key << "\t\tENTq\t" << alpha << "\t" << beta << "\t" << 0 << "\t" << ply << endl;
#endif

	if(ply > qdepth) qdepth = ply;
	if(ply >= MAXPLY-2) return beta;
    pvl[ply] = ply;
    
	if(fifty>3){ 
        if(board.gamestack.rep()) return 0;
		if(fifty >= move50) return 0;
        
	}

	// time check.  @100knps time check every 1/12 sec
    if((qnodes & 8191) == 8191)
        if(clock() >= st /* || Bioskey() */){
            stopsearch = 1;
            return 0;
        }

	cMOVE bm;
    int val;
    
    tree.killer1[ply+1].to = 0x88; // next ply killer is blank for 1st move, but >1 have a killer
	tree.killer2[ply+1].to = 0x88;
	tree.hashmove[ply].to = 0x88;

#ifdef HASHING	

	cHASHENTRY * he = hashtable.probe();
    if(he != NULL){
        tree.hashmove[ply].from = he->bm.from;
        tree.hashmove[ply].to = he->bm.to;
        tree.hashmove[ply].mtype = he->bm.type;
        val = he->value;
		
		if(ABS(val) > MATE-300){
            if(val > 0) val -= (ply-0);
            else val += (ply-0);
        }

        switch(he->data & 3){
            case FAIL_LOW_UPPER: // All searched. val has best score
				if(val <= alpha){
					hashtable.stored += ((he->data & 60) != hashtable.age);
					return val;
				}
                break; 
            case FAIL_HIGH_LOWER: // Few searched. val has best
                if(val >= beta){
					hashtable.stored += ((he->data & 60) != hashtable.age);
					return val;
				}
                break;
            case EXACT_PV:
                if(val > alpha && val < beta)
                    extract(&tree.hashmove[ply]);
                hashtable.stored += ((he->data & 60) != hashtable.age);
				return val;
                break;
        }
    }
	
#endif

	bm = tree.hashmove[ply];

	bool incheck = board.incheck();
	int nt = FAIL_LOW_UPPER;

	val = board.evaluate();

	if(val > alpha){
		if(val >= beta){
			hashtable.store(0, val, FAIL_HIGH_LOWER, &bm);

#ifdef TRACEc
			logfile << key << "\t\tqSTP\t" << alpha << "\t" << beta << "\t" << val << "\t" << 0 << "\t" << ply << endl;
#endif				
			return val;
		}
		alpha = val;
	}


	tree.gencap();

	int legal = 0, stdpat = val, best = -NFINITY;

	//int matplus = board.material[side]-board.material[xside]+max_pos_value;

	cMOVE * cm = tree.nextmove[ply-1];
    while(cm < tree.nextmove[ply]){
        tree.isortone(cm);

		if((!incheck // only caps were generated, so they're all 'good'
			
#ifdef FUTILITYq
			&& stdpat + max_pos_gain + matval[board.square[cm->to].ptype] > alpha
#endif
			)
		|| (incheck && cm->is_pseudo_evasion()) // consider only moves which may evade
			
		){ 
			
#ifdef TRACEc
			i64 lock = key;
			cm->to_san(buffer);
#endif
			
			if(board.domove(cm)){
				legal++;

#ifdef TRACEc
	logfile << lock << "\t" << buffer << "\tCSE\t" << -alpha << "\t" << -beta << "\t" << 0 << "\t" << ply << endl;;
#endif
				val = -qsearch(-beta, -alpha);
				
				board.unmove();
				if(val > best){
					if(val > alpha){   
						if(val >= beta){  
							if(cm->score < 128) bump(cm, 1);
							fh1 += (legal == 1);
							fhn++;
							hashtable.store(0, val, FAIL_HIGH_LOWER, cm);
#ifdef TRACEc
							cm->to_san(buffer);
							logfile << key << "\t" << buffer << "\tcCUT\t" << alpha << "\t" << beta << "\t" << val << "\t" << 0 << "\t" << ply << endl;
#endif
							return val;
						}
						alpha = val;
						//bm = *cm;
						nt = EXACT_PV;
						/* update the PV */
						pv[ply][ply] = *cm;
						for (int j = ply + 1; j < pvl[ply + 1]; ++j)
							pv[ply][j] = pv[ply + 1][j];
						pvl[ply] = pvl[ply + 1];
					}
					bm = *cm;
					best = val;
					if(best > stdpat) stdpat = best;
				}
			}
		}
        cm++;
    }
	
	//if(nt == EXACT_PV && bm.score < 128) bump(&bm, 1);

    hashtable.store(0, stdpat, nt, &bm);

#ifdef TRACEc
	bm.to_san(buffer);
	if(nt == EXACT_PV)
		logfile << key << "\t" << buffer << "\tcPVN\t" << alpha << "\t" << beta << "\t" << stdpat << "\t" << 0 << "\t" << ply << endl;
	else
		logfile << key << "\t" << buffer << "\tcALL\t" << alpha << "\t" << beta << "\t" << stdpat << "\t" << 0 << "\t" << ply << endl;
#endif	
	
	assert(stdpat > -MATE && stdpat < MATE);

	return stdpat;
}

