#include "chess.h"
#include "data.h"
#include "search.h"
#include "weights.h"
#include "moves.h"

#ifdef _LINUX
#include <stdlib.h>
#endif

void bump(cMOVE *m, int d){
    unsigned int h = (historic[side][m->from][m->to] += d * d); // need +1 for 0 depth
    if((h >> maxhist[side]) > MAXHISTSCORE) {maxhist[side]++;}
	if(tree.killer1[ply].is_equal(m));
	else{
		tree.killer2[ply] = tree.killer1[ply];
		tree.killer1[ply] = *m;
	}
}

int pvsearch(int alpha, int beta, int depth, int extensions, int flags){
    //int val;
	assert((board.isok())==0);
	

#ifdef TRACEp
	logfile << key << "\t\tENTp\t" << alpha << "\t" << beta << "\t" << depth << "\t" << ply << endl;
#endif

	nodes++;
	
	if(ply >= MAXPLY-2) return beta;
	pvl[ply] = ply;
    
#ifdef WORSTCASE // NB worst case is I am MATED now, NOT may be MATED in 2
	int best = -MATE + (ply-1); //+ 2;
	if(best > alpha){
		if(best >= beta) return best;
		alpha = best;
	}
#else    
	int best = -NFINITY;
#endif

	// time check.  @100knps time check every 1/12 sec
    if((nodes & 8191) == 8191)
        if(clock() >= st  || Bioskey()){
            stopsearch = 1;
            return 0;
        }
    
	cMOVE bm;
	int val;
	int mt_ext = 0;
	//int ct_ext = 0;
	int bext = 0;
	int nt;
    bool incheck = board.incheck();
    
    if(flags&2); // skip all this if IIDing  cout << "!"
    else{
        if(fifty>3){ 
			if(ply>4){// 4 or more moves deep into search tree ANY repetion counts as DRAW ply *= 1  
				if(board.gamestack.rep()) return 0; // any repetition
			}
			else // otherwise, we want FULL 3 reps
				if(board.gamestack.rep2()) return 0; // only 3-fold
    		if(fifty >= move50) return 0;
        }
    	
    	
        tree.killer1[ply+1].to = 0x88;
        tree.killer2[ply+1].to = 0x88;
        tree.hashmove[ply].to = 0x88;   
        
		/*
			search inconsistencies:
			When path to position and/or bounds are different, extensions/reductions vary
			the result of search - making actual values != stored ones
		*/

#ifdef HASHING

		cHASHENTRY * he = hashtable.probe();
        if(he != NULL){
            tree.hashmove[ply].from = he->bm.from;
            tree.hashmove[ply].to = he->bm.to;
            tree.hashmove[ply].mtype = he->bm.type;
			mt_ext = he->data & 64; // this position is threatening regardless of draft
			val = he->value;
			if(ABS(val) > MATE-300){
				if(val > 0) val -= (ply-0);
				else val += (ply-0);
			}
			nt = he->data & 3;
			if(he->depth >= depth){
				switch(nt){
				case FAIL_LOW_UPPER: //	All	searched. val has best score
					if(val <= alpha){
						//hashtable.useful++;
						if((he->data & 60) != hashtable.age){
							hashtable.stored++;
							he->data &= 195;
							he->data |= hashtable.age;
						}
						return val;
					}
					// score is	not	higher than	this
					if(val < beta){
						//beta = val;	// causes inconsistencies and ruins PV
						flags &= 254; // null-move will	not	fail high, so dont try
					}
					break;
				case FAIL_HIGH_LOWER: // Few searched. val has best
					if(val >= beta){
						//hashtable.useful++;
						if((he->data & 60) != hashtable.age){
							hashtable.stored++;
							he->data &= 195;
							he->data |= hashtable.age;
						}
						return val;
					}
					// causes inconsistencies and ruins PV
					// if(val > alpha) alpha = val;
					break; 
				case EXACT_PV:
					//hashtable.useful++;
					if((he->data & 60) != hashtable.age){
						hashtable.stored++;
						he->data &= 195;
						he->data |= hashtable.age;
					}
					if(val > alpha && val <	beta){
						//pv[ply][ply] = tree.hashmove[ply];
						extract(&tree.hashmove[ply]);
					}
					return val;
					break;
				}
			}
			else if(he->depth >= depth-2){
				if(nt == FAIL_LOW_UPPER || nt == EXACT_PV)
					if(val < beta) flags &= 254;
			}
        }
    
#endif 

	}
    
#ifdef NULLMOVE    
    /*
        NULL move
        note: do not save unverified null search in hash unless it can be identified as such
    */
    //*                                           
    int reduce = 2;// + (int)(depth>6);
    if((flags&3) == 1 // DO NULL-MOVE iff:  didnt just do one,
        //&& depth > reduce+1   //  == never drop into qsearch. wont see mate
		&& board.null_ok() // >= Rook
		&& !incheck 
        && !mt_ext
		//&& extensions == 0
        //&& beta > -MATE+300  // !getting mated. -seems to find shortest mate w/this
		//&& beta < MATE-300  // unlikely to fail high  
		//&& alpha < MATE-300  // !mating [also, unlikely to fail high b>a]
		//&& depth+ply-2 <= idepth  // not extended experiment( == is or reduced!)
									// no more than 1 mate threat in line
        //&& beta != alpha+1 // exclude pv AND null-move null-window searches ???

		){

#ifdef TRACEp
	logfile << key << "\t\tNUL\t" << -beta << "\t" << -beta+1 << "\t" << depth << "\t" << ply << endl;
#endif

        tree.nextmove[ply] = tree.nextmove[ply-1];
        board.gamestack.cGAMEHIST::pushnull();
        
		key ^= zenpassant[enpassant];//extract old ep
		enpassant = 0;
		key ^= zenpassant[enpassant];// apply new ep
		key = ~key;

        side ^= 1;
        xside ^= 1;
        
        ply++;
        /*
            a 3rep or fifty-move draw may cause cut-off
        */
        fifty++; // Using 0 misses reps in history. But ++ makes the 50m rule apply after
		move50 += 100; // prevents hitting fifty after null
		
		if(depth-reduce-1 > 0)
			val = -pvsearch(-beta, -beta+1, depth-reduce-1, extensions, (flags&252)); // reset null && iid, keep verify
		else{
			nodes++;
			val = -csearch(-beta, -beta+1); // NOT! doing checks 1st ply
		}
		move50 -= 100;
		
		board.gamestack.cGAMEHIST::popnull();
		
		side ^= 1;
        xside ^= 1;
        ply--;
        if(stopsearch) return 0;
        /*
            NULL MOVE [and its peculiarities]
			1. NOT 'real' possibility. therefore...?
            2. ONLY a bound dependant on move order [path?]
            3. It can 'appear' to be best choice! [zugzwang]
            4. Inconsistent due to changing 'conditions'
            5. If we can hash NM scores, [not sure we can] use the score, or beta?
                depth,  or depth-R ?
            6. Should mate threats include all MATED scores, or just MATED in 1 ?
            7. Convert mate scores to MATE-300? ONLY IF WE USE VAL!
			8. Raise alpha to val? ie, if alpha < nullmove score surely we can do better
			9. shallower search can be blind to tactics
			note: mated_NOW = -MATE+(ply-1)
        */
        
		//if(val > MATE-300) val = MATE-300;

		if(val >= beta){
            tot_nm_red++;
			
			
#ifdef NUL_RET_VAL
            //hashtable.store(depth, val, FAIL_HIGH_LOWER, &no_move);
			return val;
#else
            //hashtable.store(depth, beta, FAIL_HIGH_LOWER, &no_move);
			return beta; //val; // do not know???-reduce
#endif
        }
        
		if(val == -MATE+(ply-1)+2){//is much better than < -MATE+300){ //
            mt_ext = 64;
		}

		/*
			doing null after a non-cap which results in material loss
			indicates a threat which *may* be a pointless delay/horizon effect move
		*/
#ifdef CAP_THREAT
		else if(board.gamestack.stack[board.gamestack.hply-1].matscore == board.material[xside]-board.material[side]){
			if(val < alpha-50){ // alpha > 0 && 
			//if(val < board.material[side]-board.material[xside]-300)
				ct_ext = 1;
			}
		}
#endif
		//if(val > alpha) alpha = val; // we dont trust val
    }
    //*/
#endif

	//NOTE: if null-move-flag is 0, parent was null move 
	//		so, were doing shallow, null-window search, for any fail-high
	//      maybe we dont need to do iid - besides, beta == alpha+1
	//		so how well will iid work anyway?
	bool needed = true;

#ifdef IID  
	if(depth > 2 && tree.hashmove[ply].to == 0x88) do{  //beta != alpha+1 &&  (alpha != beta-1)
		
		if(ply & 1){ // same as root
			if(alpha != rootalpha || beta != rootbeta) break;
		}else        // alpha = -beta  beta = -alpha
			if(-beta != rootalpha || -alpha != rootbeta) break;

#ifdef TRACEp
	logfile << key << "\tIID\t" << alpha << "\t" << beta << "\t" << depth << "\t" << ply << endl;
#endif

		val = pvsearch(alpha, beta, depth - 2, extensions, (flags&254)|2); // preserve verify, reset Null, set IID
		if(stopsearch) return 0;

#ifndef TRACEp		
        if(val <= alpha)
            val = pvsearch(-NFINITY, alpha+1, depth - 2, extensions, (flags&254)|2);// null isnt a move. set iid flag
		//* this is a little better
		else if(val >= beta)
            val = pvsearch(beta-1, NFINITY, depth - 2, extensions, (flags&254)|2);// null isnt a move. set iid flag
		//*/
		if(stopsearch) return 0;
#endif
		tree.hashmove[ply] = pv[ply][ply];
		needed = false;
	}while(0);
#endif
    
	
	
    if(needed) tree.gen();

	
	/*
		WHOLE NODE 'THREAT' EXTENSIONS
		
	*/
	
	int legal, ext;
	int extended = 0;//(int)(ply > idepth*2);///(depth > 3);//(int)(extensions > idepth); //(int)


	if(incheck){
		legal = tree.filterlegals();
		if(legal == 0){
			best = -MATE+(ply-1);
			hashtable.store(depth, best, EXACT_PV, &no_move);
			return best;
		}
#ifdef ONEREPLY
		// one reply to check. indicative of being in a 'tight' situation
		if(legal == 1
			//&& !extended
			//&& beta > -MATE+300  // not _already_ getting mated
			//&& extensions < (idepth)
			){ 
				tot_1reply++;
				depth++;
				extended = 1;
		}
#endif
	}

#ifdef NULLMOVE	
	// mate threat. indicative of dire threat 'looming'. exlusive??? 
	if(mt_ext){
		if(!extended
			//&& depth < 3
			//&& beta > -MATE+300  // not _already_ getting mated
			//&& (extensions & (2<<side)) == 0
			//&& extensions < (idepth)
		){ 
			tot_mt_ext++;
			depth++;
			extended = 1; //(int)(ply+depth-2 > idepth * 4); //1; //  experimental
			//extensions++;
		}
	}//else extensions &= ~(1<<side);
    
#endif

	//extended |= (int)(ply > idepth*2);
	
	nt = FAIL_LOW_UPPER;
	
	//best = -NFINITY;
	
	flags &= 253; // reset iid and 
    //flags |= 1;   // default to null_move for subsequent trees. doing later
	legal = 0;
	//int check_ext;
				
    cMOVE * cmove = tree.nextmove[ply-1];
	while(cmove < tree.nextmove[ply]){
		tree.isortone(cmove);

#ifdef TRACEp
		cmove->to_san(buffer);
		i64 lock = key;
#endif

		if(board.domove(cmove)){
			legal++;
			ext = 0;
			//check_ext = extensions;
			// MOVE EXTENSIONS

			if(extended);// whole node or we simply shut down ext for too many
#ifdef RECAP			
			//*/ recap extension  
			else if(board.gamestack.was_recap()
				//&& extensions == 0
				//&& beta > 0
				//&& alpha > 0
				){//    
				tot_recap_ext++;
				//extensions = 1;
				ext = 1;
			}//*/
#endif
#ifdef PAWN27
			// pawn to 7th
			else if(board.gamestack.was_p27()){
				tot_p7_ext++;
				ext = 1;
			}
#endif
#ifdef CHECKING
			// check extension
			else if(board.incheck() 
				){
				//if(alpha < MATE-300){ // POSSIBLY A PROBLEM
					//check_ext++;
					tot_ck_ext++;
					ext = 1;
				//}//else check_ext = 0;
			}
#endif
		
#ifdef FUTILITY
			/*
				No, this isnt actually futility pruning, but rather lazy eval
				eccept here, we know we are in check or giving check and can
				avoid it. plus save a call to quies() as well as to eval()
				we skip this if we're already reduced by nullmove.
				max_pos_value is exactly that - there is no way the full eval would fail
			*/

			else if(depth == 1 
				&& !incheck
				// && flags&1 // parent was NOT null-move - use this if using isearch
				){
					val = board.material[xside]-board.material[side]+max_pos_value;
					if(val < alpha){
						/*
						val = -csearch(-beta, -alpha);
						if(val > alpha){
							cout << "oops " << val << " " 
							<< board.material[xside]-board.material[side]
							<< endl;
						}//*/
						tot_fut_cut++;
						goto skip;
					}
			}
#endif			
			
			
			
#ifdef TRACEp
	logfile << lock << "\t" << buffer << "\t";
#endif
			
			
			if(nt == EXACT_PV){ //best > -NFINITY){ //

				if(depth+ext-1 > 0){

#ifdef TRACEp
	logfile << "PVS\t" << -alpha-1 << "\t" << -alpha << "\t" << depth << "\t" << ply << endl;
#endif

					val = -pvsearch(-alpha-1, -alpha, depth+ext-1, 0, flags|1);
				}
				else{

#ifdef TRACEp
	logfile << "PVSi\t" << -alpha-1 << "\t" << -alpha << "\t" << depth << "\t" << ply << endl;
#endif
					nodes++;
					val = -csearch(-alpha-1, -alpha);
				}
				if(!stopsearch){// && val > alpha){
					if(val > alpha && val < beta){
						if(depth+ext-1 > 0){

#ifdef TRACEp
	logfile << lock << "\t" << buffer << "\tRES\t" << -beta << "\t" << -alpha << "\t" << depth << "\t" << ply << endl;
#endif

							val = -pvsearch(-beta, -alpha, depth+ext-1, 0, flags|1);
						}
						else{
#ifdef TRACEp
	logfile << lock << "\t" << buffer << "\tRESi\t" << -beta << "\t" << -alpha << "\t" << depth << "\t" << ply << endl;
#endif
							nodes++;
							val = -csearch(-beta, -alpha);
						}
					}
				}
			}
			else{ 
				if(depth+ext-1 > 0){

#ifdef TRACEp
	logfile << "ONE\t" << -beta << "\t" << -alpha << "\t" << depth << "\t" << ply << endl;
#endif

					val = -pvsearch(-beta, -alpha, depth+ext-1, 0, flags|1);
				}
				else{
#ifdef TRACEp
	logfile << "ONEi\t" << -beta << "\t" << -alpha << "\t" << depth << "\t" << ply << endl;
#endif
					nodes++;
					val = -csearch(-beta, -alpha);
				}
			}

//skip:
			board.unmove();

			if(stopsearch) return 0;

			if(val > best){
				if(val > alpha){  
					if(val >= beta){ 
						if(cmove->score < 128) bump(cmove, depth+1);
						fh1 += (legal == 1);
						fhn++;
						//tree.hashmove[ply] = *cmove; // for iid 
						hashtable.store(depth, val, FAIL_HIGH_LOWER | mt_ext, cmove);
						return val;//beta;//
					}
					alpha = val;
					nt = EXACT_PV;

					/* update the PV */
					pv[ply][ply] = *cmove;  
					for (int j = ply + 1; j < pvl[ply + 1]; ++j)
						pv[ply][j] = pv[ply + 1][j];
					pvl[ply] = pvl[ply + 1];

				}
				bm = *cmove;
				bext = ext; // remember extension status for later
				best = val;
			}
		}
		cmove++;
	}
    if(legal == 0){
        best = 0;
        nt = EXACT_PV;
		//tree.hashmove[ply].from = 0x88;// dont let iid or hash have a good move
        hashtable.store(depth, best, nt, &no_move);
        return best;
    }

#ifdef CAP_THREAT
	/*
		If not moving, after a non-cap results in material loss eg, pawn threatens rook;
		How can we tell if the threat was empty - pushing something over horizon?
		- this can only hide 'positional' losses - not material ones, so if we're failing
		low, we can assume this isnt just a delay tactic, but we may want to see if we can
		do better with the score so it beats out a smaller variation elseswhere
		- can happen whether winning or losing, so alpha>0 no help
		- we be failing low BECAUSE xside 
	*/
	if(ct_ext
		&& !extended // if we were in check, no null move could have been tried
		&& bext == 0 // not already extended
		//&& alpha > 0
		&& nt == EXACT_PV
		){
			tot_delay++;
			board.domove(&bm);
			val = -pvsearch(-beta, -alpha, depth, 1, flags&254);
			board.unmove();
			if(stopsearch) return 0;
			if(val > alpha){
				//if(val > best+30){
					good_delay++;
					/*
					board.display();
					cout << side << endl;
					bm.print();
					getchar();//*/
				//}
				best = val; // new best
				//if(val < beta){
					nt = EXACT_PV;
					/* update the PV */
					pv[ply][ply] = bm;  
					for (int j = ply + 1; j < pvl[ply + 1]; ++j)
						pv[ply][j] = pv[ply + 1][j];
					pvl[ply] = pvl[ply + 1];
				//}else nt = FAIL_HIGH_LOWER;
			}
		}
#endif	
    
	//tree.hashmove[ply] = bm; // for iid
	hashtable.store(depth, best, nt | mt_ext, &bm);//mtype > PROMOXN
    //if(nt > FAIL_LOW_UPPER && bm.score < 128) bump(&bm, depth+1);

	//assert(best > -MATE +(ply-1) && best < MATE - (ply-1));

#ifdef TRACEp
	bm.to_san(buffer);
	if(nt == EXACT_PV)
		logfile << key << "\t" << buffer << "\tpPVN\t" << alpha << "\t" << beta << "\t" << best << "\t" << depth << "\t" << ply << endl;
	else
		logfile << key << "\t" << buffer << "\tpALL\t" << alpha << "\t" << beta << "\t" << best << "\t" << depth << "\t" << ply << endl;
#endif

    return best;
}

