#include "chess.h"
#include "data.h"

#ifdef _LINUX
#include <stdlib.h>
#endif

    void cGAMEHIST::push(cMOVE * move, cSQUARE victim, unsigned int affected){
        stack[hply].move = *move;
        stack[hply].victim = victim;
        stack[hply].fifty = fifty;
        stack[hply].castle = castle;
        stack[hply].enpassant = enpassant;
        stack[hply].key = key;
        stack[hply].affected = affected;
		stack[hply].matscore = board.material[side]-board.material[xside];
        //cout << hply << " ";
        hply++;
    }
    /*
		what is not altered now by null moving, will be saved by next move
		and restored there
	*/
    void cGAMEHIST::pushnull(){
        stack[hply].fifty = fifty; // we change this
        stack[hply].castle = castle; // this may get changed AND RESTORED afterwards
        stack[hply].enpassant = enpassant; // we change this
        stack[hply].key = key;
		stack[hply].move.to = OFFBOARD; // so re-cap doesnt think destinations match
		stack[hply].victim.ptype = EMPTY;
        stack[hply].matscore = board.material[side]-board.material[xside];
        hply++;
    }
    
    void cGAMEHIST::pop(cMOVE &move, cSQUARE &victim, unsigned int &affected){
        hply--;
        move = stack[hply].move;
        victim = stack[hply].victim;
        fifty = stack[hply].fifty;
        castle = stack[hply].castle;
        enpassant = stack[hply].enpassant;
        key = stack[hply].key;
        affected = stack[hply].affected;
    }
    
    void cGAMEHIST::popnull(){
        hply--;
        key = stack[hply].key;
        fifty = stack[hply].fifty;
        castle = stack[hply].castle;
        enpassant = stack[hply].enpassant;
    }
    
    cMOVE * cGAMEHIST::top(){
        return &stack[hply-2].move;
    }
    //*
	int cGAMEHIST::last_victim(){
        return (matval[stack[hply-1].victim.ptype]);
    }
	//*/
	// recap: two moves to same square restoring material balance
	bool cGAMEHIST::was_recap(){
		return (stack[hply-1].move.to == stack[hply-2].move.to
			//&& matval[stack[hply-1].victim.ptype] == matval[board.square[stack[hply-1].move.to].ptype]
			&& stack[hply-2].matscore == board.material[side]-board.material[xside]
        );
	}

	bool cGAMEHIST::was_p27(){
		int dest = stack[hply-1].move.to;
		return (p27[dest] == 1 && board.square[dest].ptype == PAWN);
	}
/*    
    fifty = reversible moves done, ie, =0 after non reversible move
    key is stored at hply prior to move, then:
        hply++ , side^=1 , fifty++ [or fifty = 0]
    therefore, after move, slot hply-1 has previous position [of [now]xside]
    current position key [resulting from LAST move] is not yet stored
    position resulting from last non rev move at slot hply-fifty **TRUE** 7/24 dw
    but is not stored until just before execution of 1st reversible move
        eg., after capture, hply++, and fifty=0  : position not yet stored
        before reversible move, position stored at hply
        after reversible move, hply++, and fifty++ - previous position
            stored at (hply-1) - (fifty-1) = hply-fifty
    last 'real' game position at slot hply-ply+1

	Before execution of move on internal board, position is stored at hply,
	then hply++, fifty++. So 

	hply	fifty	-notes-		move	hply'	fifty'	ply	hply-fifty	hply-ply+1
	1		0					e2e4	2		0		-		2
	2		0					e7e5	3		0		-		3			
	3		0					g1f3	4		1		-		3			
	4		1					b8c6	5		2		-		3			
	5		2					f1b5	6		3		-		3			
	6		3		(a) (b)		d7d6	7		0		2		7			6
	7		0		(c)			b1c3	8		1		3		7			6
	8		1					g8f6	9		2		4		7			6
	9		2					c3b1	10		3		5		7			6
	10		3		    		f6g8	11		4		6		7			6
	11		4		(c')		
    
    a - real game position prior to,
	b - 1st move of search
	c - position to be repeated in search
*/    
    bool cGAMEHIST::rep(){
        //int r = 0;
        int i = hply-4;
        //int tcastle = castle & 15;
        //if(key != board.hash()) exit(1);
        while(i >= hply-fifty){
            if(key == stack[i].key 
                //&& tcastle == (stack[i].castle & 15)
                //&& enpassant == stack[i].enpassant
				) return true;    // or is start position
            i -= 2;
        }
        return false;
    }
	
	bool cGAMEHIST::rep2(){
        int r = 0;
        int i = hply-4;
        //int tcastle = castle & 15;
        //if(key != board.hash()) exit(1);
        while(i >= hply-fifty){
            if(key == stack[i].key 
                //&& tcastle == (stack[i].castle & 15)
                //&& enpassant == stack[i].enpassant
				){
                    if(i >= hply-ply+1) // if pos exists only in search
                        return true;    // or is start position
                    r++;
                    if(r > 1) return true;
            }
            i -= 2;
        }
        return false;
    }

	int cGAMEHIST::progress(int score){
		int p = 0, i = hply;
		while((i -= 2) > hply - 20){
			p += ABS(score - stack[i].move.score);
			score = stack[i].move.score;
		}
		return (p);
	}
