
//*
#include "chess.h"
#include "data.h"

void cBOARD::unmove(){
    int from, dest, color, ptype, pindex;
    cMOVE move;
    cSQUARE victim;
    unsigned int affected;
    gamestack.pop(move, victim, affected);
    from = move.from;
    dest = move.to;
    ptype = square[dest].ptype;
    color = square[dest].color;
    int other = color ^ 1;
    pindex = square[dest].pindex;
    unsigned int vbit;
    if(victim.pindex < 16) vbit = ~(P2BIT(victim.pindex, other));
    else vbit = 0xffffffff;
    side ^= 1;
    xside ^= 1;
    ply--;
    
    //clraffected(dest, color, ptype, P2BIT(pindex, color));

// CAPTURE, ENPASSANT, PROMOXQ,  
// PROMO2Q,  
// CASTLEK, CASTLEQ,
// NORMAL, SINGLE, PPUSH2,     
// PROMOXR, PROMOXB, PROMOXN,
// PROMO2R, PROMO2B, PROMO2N,
// NOMOVE
  
    switch(move.mtype){
        case CAPTURE:
                clear_all(affected & vbit);
                material[other] += matval[victim.ptype];
				force[other] += forceval[victim.ptype];
                //pcount[other][victim.ptype]++;
                square[from] = square[dest];
                pieces[color][pindex] = from;  // change address
                square[dest] = victim;
                pieces[other][victim.pindex] = dest;
                break;
        case ENPASSANT:
                clear_all(affected & vbit);
                material[other] += matval[victim.ptype];
                //pcount[other][victim.ptype]++;
                //moveto(dest, from);
                square[from] = square[dest];
                square[dest].empty();
                pieces[color][pindex] = from;
                if(side == WHITE){
                    square[dest+16] = victim;
                    pieces[other][victim.pindex] = dest+16;
                }
                else{
                    square[dest-16] = victim;
                    pieces[other][victim.pindex] = dest-16;
                }
                break;
        case PROMOXQ:
                clear_all(affected & vbit);
                material[color] += PAWN_VAL - QUEEN_VAL;
                material[other] += matval[victim.ptype];
                force[color] -= forceval[QUEEN];
				force[other] += forceval[victim.ptype];
                //pcount[color][PAWN]++;
                //pcount[color][QUEEN]--;
                //pcount[other][victim.ptype]++;
                square[from] = square[dest];
                square[from].ptype = PAWN;
                pieces[color][pindex] = from;  // change address
                square[dest] = victim;
                pieces[other][victim.pindex] = dest;
                break;
        case PROMO2Q:
                clear_all(affected);
                material[color] += PAWN_VAL - QUEEN_VAL;
                force[color] -= forceval[QUEEN];
                //pcount[color][PAWN]++;
                //pcount[color][QUEEN]--;
                square[from] = square[dest];
                square[from].ptype = PAWN;
                pieces[color][pindex] = from;  // change address
                square[dest].empty();
                break;
        case CASTLEK:
                clear_all(affected);
                if(side == WHITE){
                    //moveto(dest, from); // move king
                    square[from] = square[dest];
                    square[dest] = victim;
                    pieces[color][pindex] = from;
                    //moveto(117, 119);  // move rook
                    square[119] = square[117];
                    square[117] = victim;
                    pieces[color][square[119].pindex] = 119;
                }
                else{
                    //moveto(dest, from); // move king
                    square[from] = square[dest];
                    square[dest] = victim;
                    pieces[color][pindex] = from;
                    //moveto(5, 7); // move rook
                    square[7] = square[5];
                    square[5] = victim;
                    pieces[color][square[7].pindex] = 7;
                }
                break;
        case CASTLEQ:
                clear_all(affected);
                if(side == WHITE){
                    //moveto(dest, from); // move king
                    square[from] = square[dest];
                    square[dest] = victim;
                    pieces[color][pindex] = from;
                    //moveto(115, 112);  // move rook
                    square[112] = square[115];
                    square[115] = victim;
                    pieces[color][square[112].pindex] = 112;
                }
                else{
                    //moveto(dest, from); // move king
                    square[from] = square[dest];
                    square[dest] = victim;
                    pieces[color][pindex] = from;
                    //moveto(3, 0); // move rook
                    square[0] = square[3];
                    square[3] = victim;
                    pieces[color][square[0].pindex] = 0;
                }
                break;
        case NORMAL:
        case SINGLE:
                clear_all(affected);
                //moveto(dest, from);
                square[from] = square[dest];
                square[dest] = victim;
                pieces[color][pindex] = from;
                break;
        case PPUSH2:
                clear_all(affected);
                //moveto(dest, from);
                square[from] = square[dest];
                square[dest] = victim;
                pieces[color][pindex] = from;
                break;
        case PROMOXR:
                clear_all(affected & vbit);
                material[color] += PAWN_VAL - ROOK_VAL;
                material[other] += matval[victim.ptype];
                force[color] -= 5;
				force[other] += forceval[victim.ptype];
                //pcount[color][PAWN]++;
                //pcount[color][ROOK]--;
                //pcount[other][victim.ptype]++;
                square[from] = square[dest];
                square[from].ptype = PAWN;
                pieces[color][pindex] = from;  // change address
                square[dest] = victim;
                pieces[other][victim.pindex] = dest;
                break;
        case PROMOXB:
                clear_all(affected & vbit);
                material[color] += PAWN_VAL - BISHOP_VAL;
                material[other] += matval[victim.ptype];
                force[color] -= 3;
				force[other] += forceval[victim.ptype];
                //pcount[color][PAWN]++;
                //pcount[color][BISHOP]--;
                //pcount[other][victim.ptype]++;
                square[from] = square[dest];
                square[from].ptype = PAWN;
                pieces[color][pindex] = from;  // change address
                square[dest] = victim;
                pieces[other][victim.pindex] = dest;
                break;
        case PROMOXN:
                clear_all(affected & vbit);
                material[color] += PAWN_VAL - KNIGHT_VAL;
                material[other] += matval[victim.ptype];
                force[color] -= 3;
				force[other] += forceval[victim.ptype];
                //pcount[color][PAWN]++;
                //pcount[color][KNIGHT]--;
                //pcount[other][victim.ptype]++;
                square[from] = square[dest];
                square[from].ptype = PAWN;
                pieces[color][pindex] = from;  // change address
                square[dest] = victim;
                pieces[other][victim.pindex] = dest;
                break;
        case PROMO2R:
                clear_all(affected);
                material[color] += PAWN_VAL - ROOK_VAL;
                force[color] -= 5;
                //pcount[color][PAWN]++;
                //pcount[color][ROOK]--;
                square[from] = square[dest];
                square[from].ptype = PAWN;
                pieces[color][pindex] = from;  // change address
                square[dest].empty();
                break;
        case PROMO2B:
                clear_all(affected);
                material[color] += PAWN_VAL - BISHOP_VAL;
                force[color] -= 3;
                //pcount[color][PAWN]++;
                //pcount[color][BISHOP]--;
                square[from] = square[dest];
                square[from].ptype = PAWN;
                pieces[color][pindex] = from;  // change address
                square[dest].empty();
                break;
        case PROMO2N:
                clear_all(affected);
                material[color] += PAWN_VAL - KNIGHT_VAL;
                force[color] -= 3;
                //pcount[color][PAWN]++;
                //pcount[color][KNIGHT]--;
                square[from] = square[dest];
                square[from].ptype = PAWN;
                pieces[color][pindex] = from;  // change address
                square[dest].empty();
                break;
        case NOMOVE:
                break;
    }
    
    //setaffected(from, color, ptype, P2BIT(pindex, color));
    
	set_all(affected);
    
	assert(key == hash());

	/*if(!isok()){
		display();
		puts("after unmove:\n");
		cout << material[WHITE] << " " << material[BLACK] << endl;
		getchar();
	}*/
    //if(key != board.hash()) cout << "\a\a";
}    
//*/
